from typing import cast

from draive.evaluation import EvaluationScore, EvaluationScoreValue, evaluator
from draive.multimodal import Multimodal, MultimodalContent, MultimodalTagElement
from draive.stages import Stage

__all__ = ("consistency_evaluator",)


INSTRUCTION: str = """\
You are evaluating the provided content according to the defined criteria.

<INSTRUCTION>
Compare the REFERENCE and the EVALUATED content by carefully examining them, then rate\
 the EVALUATED content using solely a consistency metric according to the EVALUATION_CRITERIA.
Think step by step and provide explanation of the score before the final score.
Use the explained RATING scale and the requested FORMAT to provide the result.
</INSTRUCTION>

<EVALUATION_CRITERIA>
Evaluated metric is consistency - a factual alignment between the REFERENCE and the EVALUATED content.
A factually consistent content contains only elements that are entailed by the REFERENCE content.
</EVALUATION_CRITERIA>
{guidelines}
<RATING>
Assign a consistency score using exact name of one of the following values:
- "poor" is very low consistency, the content contains multiple hallucinated facts\
 or significant misalignments with the reference content.
- "fair" is low consistency, the content has several instances of information not supported by\
 the reference content.
- "good" is moderate consistency, the content is mostly consistent but contains a few unsupported\
 statements.
- "excellent" is high consistency, the content is largely consistent with minor discrepancies.
- "perfect" is very high consistency, the content is fully consistent with the reference content,\
 containing only supported information.
Use the "none" value for content that cannot be rated at all.
</RATING>

<FORMAT>
The final result containing only the numerical score value HAVE to be put inside a `RESULT`\
 xml tag within the result i.e. `<RESULT>good</RESULT>`.
</FORMAT>
"""  # noqa: E501


@evaluator(name="consistency")
async def consistency_evaluator(
    evaluated: Multimodal,
    /,
    *,
    reference: Multimodal,
    guidelines: str | None = None,
) -> EvaluationScore:
    if not evaluated:
        return EvaluationScore(
            value=0,
            comment="Input was empty!",
        )

    if not reference:
        return EvaluationScore(
            value=0,
            comment="Reference was empty!",
        )

    completion: MultimodalContent = await Stage.completion(
        MultimodalContent.of(
            "<REFERENCE>",
            reference,
            "</REFERENCE>\n<EVALUATED>",
            evaluated,
            "</EVALUATED>",
        ),
        instruction=INSTRUCTION.format(
            guidelines=f"\n<GUIDELINES>\n{guidelines}\n</GUIDELINES>\n"
            if guidelines is not None
            else ""
        ),
    ).execute()

    if result := MultimodalTagElement.parse_first(
        "RESULT",
        content=completion,
    ):
        return EvaluationScore.of(
            cast(EvaluationScoreValue, result.content.to_str()),
            comment=completion.to_str(),
        )

    else:
        raise ValueError("Invalid evaluator result:\n%s", completion)



from collections.abc import Mapping, Sequence
from types import EllipsisType, NoneType, UnionType

from haiway.state import AttributeAnnotation

from draive import DataModel
from draive.parameters import ParametersSpecification
from draive.parameters.specification import parameter_specification


def test_specifications() -> None:
    assert parameter_specification(
        AttributeAnnotation(origin=str),
        description=None,
    ) == {
        "type": "string",
    }
    assert parameter_specification(
        AttributeAnnotation(origin=int),
        description=None,
    ) == {
        "type": "integer",
    }
    assert parameter_specification(
        AttributeAnnotation(origin=float),
        description=None,
    ) == {
        "type": "number",
    }
    assert parameter_specification(
        AttributeAnnotation(origin=bool),
        description=None,
    ) == {
        "type": "boolean",
    }
    assert parameter_specification(
        AttributeAnnotation(origin=NoneType),
        description=None,
    ) == {
        "type": "null",
    }
    assert parameter_specification(
        AttributeAnnotation(
            origin=Sequence,
            arguments=[
                AttributeAnnotation(origin=str),
            ],
        ),
        description=None,
    ) == {"type": "array", "items": {"type": "string"}}
    assert parameter_specification(
        AttributeAnnotation(
            origin=tuple,
            arguments=[
                AttributeAnnotation(origin=str),
                AttributeAnnotation(origin=EllipsisType),
            ],
        ),
        description=None,
    ) == {"type": "array", "items": {"type": "string"}}
    assert parameter_specification(
        AttributeAnnotation(
            origin=tuple,
            arguments=[
                AttributeAnnotation(origin=str),
                AttributeAnnotation(origin=str),
            ],
        ),
        description=None,
    ) == {"type": "array", "prefixItems": [{"type": "string"}, {"type": "string"}]}
    assert parameter_specification(
        AttributeAnnotation(
            origin=Sequence,
            arguments=[
                AttributeAnnotation(origin=str),
            ],
        ),
        description=None,
    ) == {"type": "array", "items": {"type": "string"}}
    assert parameter_specification(
        AttributeAnnotation(
            origin=Mapping,
            arguments=[
                AttributeAnnotation(origin=str),
                AttributeAnnotation(origin=str),
            ],
        ),
        description=None,
    ) == {"type": "object", "additionalProperties": {"type": "string"}}
    assert parameter_specification(
        AttributeAnnotation(
            origin=UnionType,
            arguments=[
                AttributeAnnotation(origin=str),
                AttributeAnnotation(origin=int),
            ],
        ),
        description=None,
    ) == {"oneOf": [{"type": "string"}, {"type": "integer"}]}


def test_basic_specification() -> None:
    class TestModel(DataModel):
        str_value: str
        int_value: int
        float_value: float
        bool_value: bool
        none_value: None
        list_value: Sequence[str]
        dict_value: Mapping[str, str]

    specification: ParametersSpecification = {
        "type": "object",
        "properties": {
            "str_value": {"type": "string"},
            "int_value": {"type": "integer"},
            "float_value": {"type": "number"},
            "bool_value": {"type": "boolean"},
            "none_value": {"type": "null"},
            "list_value": {"type": "array", "items": {"type": "string"}},
            "dict_value": {"type": "object", "additionalProperties": {"type": "string"}},
        },
        "required": [
            "str_value",
            "int_value",
            "float_value",
            "bool_value",
            "none_value",
            "list_value",
            "dict_value",
        ],
    }
    assert TestModel.__PARAMETERS_SPECIFICATION__ == specification


def test_parametrized_specification() -> None:
    class TestModel[Param](DataModel):
        param: Param

    assert TestModel.__PARAMETERS_SPECIFICATION__ == {
        "type": "object",
        "properties": {
            "param": {
                "type": "object",
                "additionalProperties": True,
            },
        },
        "required": ["param"],
    }
    assert TestModel[str].__PARAMETERS_SPECIFICATION__ == {
        "type": "object",
        "properties": {
            "param": {"type": "string"},
        },
        "required": ["param"],
    }
    assert TestModel[int].__PARAMETERS_SPECIFICATION__ == {
        "type": "object",
        "properties": {
            "param": {"type": "integer"},
        },
        "required": ["param"],
    }
    assert TestModel[str] == TestModel[str]
    assert TestModel[int] == TestModel[int]
    assert TestModel[str] != TestModel[int]


def test_nested_parametrized_specification() -> None:
    class TestModelNested[Param](DataModel):
        param: Param

    class TestModelHolder[Param: DataModel](DataModel):
        param: Param

    class TestModel[Param: DataModel](DataModel):
        param: TestModelHolder[Param]

    assert TestModel.__PARAMETERS_SPECIFICATION__ == {
        "type": "object",
        "properties": {
            "param": {
                "type": "object",
                "properties": {
                    "param": {
                        "type": "object",
                        "additionalProperties": True,
                    },
                },
                "required": ["param"],
            },
        },
        "required": ["param"],
    }
    assert TestModel[TestModelNested[str]].__PARAMETERS_SPECIFICATION__ == {
        "type": "object",
        "properties": {
            "param": {
                "type": "object",
                "properties": {
                    "param": {
                        "type": "object",
                        "properties": {
                            "param": {
                                "type": "string",
                            },
                        },
                        "required": ["param"],
                    }
                },
                "required": ["param"],
            }
        },
        "required": ["param"],
    }
    assert TestModel[TestModelNested[int]].__PARAMETERS_SPECIFICATION__ == {
        "type": "object",
        "properties": {
            "param": {
                "type": "object",
                "properties": {
                    "param": {
                        "type": "object",
                        "properties": {
                            "param": {
                                "type": "integer",
                            },
                        },
                        "required": ["param"],
                    }
                },
                "required": ["param"],
            }
        },
        "required": ["param"],
    }
    assert TestModel[TestModelNested[str]] == TestModel[TestModelNested[str]]
    assert TestModel[TestModelNested[int]] == TestModel[TestModelNested[int]]
    assert TestModel[TestModelNested[str]] != TestModel[TestModelNested[int]]



from collections.abc import Iterable, Sequence
from typing import Any, Protocol, Self, runtime_checkable
from uuid import uuid4

from haiway import State

from draive.commons import Meta, MetaValue
from draive.instructions import Instruction
from draive.multimodal import Multimodal, MultimodalContent
from draive.tools import Toolbox

__all__ = (
    "ChoiceCompletion",
    "ChoiceOption",
    "SelectionException",
)


class ChoiceOption(State):
    @classmethod
    def of(
        cls,
        content: Multimodal,
        /,
        identifier: str | None = None,
        **meta: MetaValue,
    ) -> Self:
        assert identifier is None or len(identifier) > 0, "Identifier can't be empty"  # nosec: B101

        return cls(
            identifier=identifier or uuid4().hex,
            content=MultimodalContent.of(content),
            meta=Meta.of(meta),
        )

    identifier: str
    content: MultimodalContent
    meta: Meta


@runtime_checkable
class ChoiceCompletion(Protocol):
    async def __call__(
        self,
        *,
        instruction: Instruction | str,
        options: Sequence[ChoiceOption],
        input: Multimodal,  # noqa: A002
        toolbox: Toolbox,
        examples: Iterable[tuple[Multimodal, ChoiceOption]] | None,
        **extra: Any,
    ) -> ChoiceOption: ...


class SelectionException(Exception):
    pass



from typing import Literal, cast

from anthropic import AsyncAnthropic
from haiway import getenv_str

__all__ = ("AnthropicAPI",)


class AnthropicAPI:
    __slots__ = (
        "_api_key",
        "_aws_region",
        "_base_url",
        "_client",
        "_provider",
        "_timeout",
    )

    def __init__(
        self,
        provider: Literal["anthropic", "bedrock"] = "anthropic",
        /,
        *,
        base_url: str | None = None,
        api_key: str | None = None,
        aws_region: str | None = None,
        timeout: float = 60.0,
    ) -> None:
        self._provider: Literal["anthropic", "bedrock"] = provider
        self._base_url: str | None = base_url
        self._api_key: str | None = api_key or getenv_str("ANTHROPIC_API_KEY")
        self._aws_region: str | None = aws_region or getenv_str("AWS_BEDROCK_REGION")
        self._timeout: float = timeout
        self._client: AsyncAnthropic = self._prepare_client()

    def _prepare_client(self) -> AsyncAnthropic:
        match self._provider:
            case "anthropic":
                return AsyncAnthropic(
                    base_url=self._base_url,
                    api_key=self._api_key,
                    timeout=self._timeout,
                    max_retries=0,  # disable library retries
                )

            case "bedrock":
                from anthropic import AsyncAnthropicBedrock

                return cast(
                    AsyncAnthropic,  # it is not actually the same but for our purpose it works
                    AsyncAnthropicBedrock(
                        base_url=self._base_url,
                        aws_region=self._aws_region,
                        timeout=self._timeout,
                    ),
                )

    async def _initialize_client(self) -> None:
        await self._client._client.aclose()
        self._client = self._prepare_client()
        await self._client.__aenter__()

    async def _deinitialize_client(self) -> None:
        await self._client.__aexit__(
            None,
            None,
            None,
        )



from collections.abc import Sequence

from draive import State
from draive.parameters import ParameterSpecification


def test_validation() -> None:
    class ModelState(State):
        specification: Sequence[ParameterSpecification]

    specification: Sequence[ParameterSpecification] = (
        {  # union
            "oneOf": (
                {
                    "type": "integer",
                },
                {
                    "type": "string",
                    "enum": ("enum",),
                },
                {
                    "type": "null",
                },
            ),
        },
        {  # string format
            "type": "string",
            "format": "uuid",
        },
        {  # string enum
            "type": "string",
            "enum": ("enum",),
        },
        {  # string
            "type": "string",
        },
        {  # integer enum
            "type": "integer",
            "enum": (42,),
        },
        {  # integer
            "type": "integer",
        },
        {  # number enum
            "type": "number",
            "enum": (42.0,),
        },
        {  # number
            "type": "number",
        },
        {  # boolean
            "type": "boolean",
        },
        {  # null
            "type": "null",
        },
        {  # array typed
            "type": "array",
            "items": {
                "type": "integer",
            },
        },
        {  # tuple
            "type": "array",
            "prefixItems": (
                {
                    "type": "integer",
                },
            ),
        },
        {  # array any
            "type": "array",
        },
        {  # object
            "type": "object",
            "properties": {
                "integer": {
                    "type": "integer",
                }
            },
        },
        {  # dict
            "type": "object",
            "additionalProperties": {
                "type": "integer",
            },
        },
        {  # any object
            "type": "object",
            "additionalProperties": True,
        },
        {  # reference
            "$ref": "Reference",
        },
    )
    # not raises and has same value
    assert ModelState(specification=specification).specification == specification



from collections.abc import Sequence
from itertools import chain
from typing import ClassVar, Self, final, overload

from draive.commons import Meta, MetaValues
from draive.multimodal.media import MediaContent, MediaData, MediaKind, MediaReference
from draive.multimodal.meta import MetaContent
from draive.multimodal.text import TextContent
from draive.parameters import DataModel

__all__ = (
    "Multimodal",
    "MultimodalContent",
    "MultimodalContentConvertible",
    "MultimodalContentElement",
)


MultimodalContentElement = TextContent | MediaContent | MetaContent | DataModel
MultimodalContentConvertible = str | MultimodalContentElement


@final  # TODO: optimize performance
class MultimodalContent(DataModel):
    empty: ClassVar[Self]  # defined after the class

    @classmethod
    def of(
        cls,
        *elements: Self | MultimodalContentConvertible,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        match elements:
            case [MultimodalContent() as content]:
                # if we got just a single content use it as is
                return content  # pyright: ignore[reportReturnType]

            case elements:
                return cls(
                    parts=tuple(
                        _merge_texts(
                            *chain.from_iterable(
                                _extract_parts(
                                    element,
                                    meta=Meta.of(meta),
                                )
                                for element in elements
                            )
                        )
                    ),
                )

    parts: Sequence[MultimodalContentElement]

    @property
    def has_media(self) -> bool:
        return any(_is_media(part) for part in self.parts)

    def media(
        self,
        media: MediaKind | None = None,
    ) -> Sequence[MediaContent]:
        if media is None:
            return tuple(part for part in self.parts if isinstance(part, MediaContent))

        else:
            return tuple(
                part for part in self.parts if isinstance(part, MediaContent) and part.kind == media
            )

    def is_media(
        self,
        media: MediaKind | None = None,
    ) -> bool:
        if media is None:
            return all(isinstance(part, MediaContent) for part in self.parts)

        else:
            return all(isinstance(part, MediaContent) and part.kind == media for part in self.parts)

    def without_media(self) -> Self:
        return self.__class__(
            parts=tuple(part for part in self.parts if not isinstance(part, MediaContent)),
        )

    def is_artifact(self) -> bool:
        return all(_is_artifact(part) for part in self.parts)

    @property
    def has_artifacts(self) -> bool:
        return any(_is_artifact(part) for part in self.parts)

    @overload
    def artifacts(
        self,
        /,
    ) -> Sequence[DataModel]: ...

    @overload
    def artifacts[Artifact: DataModel](
        self,
        model: type[Artifact],
        /,
    ) -> Sequence[Artifact]: ...

    def artifacts[Artifact: DataModel](
        self,
        model: type[Artifact] | None = None,
        /,
    ) -> Sequence[Artifact] | Sequence[DataModel]:
        if model is None:
            return tuple(part for part in self.parts if _is_artifact(part))

        else:
            return tuple(part for part in self.parts if isinstance(part, model))

    def without_artifacts(self) -> Self:
        return self.__class__(
            parts=tuple(part for part in self.parts if not _is_artifact(part)),
        )

    def is_meta(self) -> bool:
        return all(isinstance(part, MetaContent) for part in self.parts)

    def meta(
        self,
        *,
        category: str | None = None,
    ) -> Sequence[MetaContent]:
        if category is not None:
            return tuple(
                part
                for part in self.parts
                if isinstance(part, MetaContent) and part.category == category
            )

        else:
            return tuple(part for part in self.parts if isinstance(part, MetaContent))

    def without_meta(self) -> Self:
        return self.__class__(
            parts=tuple(part for part in self.parts if not isinstance(part, MetaContent)),
        )

    def to_str(
        self,
        joiner: str | None = None,
        include_data: bool = False,
    ) -> str:
        return (joiner if joiner is not None else "").join(
            _as_string(
                element,
                include_data=include_data,
            )
            for element in self.parts
        )

    def appending(
        self,
        *parts: Self | MultimodalContentConvertible,
    ) -> Self:
        if not parts:
            return self

        if len(self.parts) == 0:
            return self.__class__(
                parts=tuple(
                    _merge_texts(*chain.from_iterable(_extract_parts(part) for part in parts))
                ),
            )

        # check the last part
        match self.parts[-1]:
            case TextContent() as text:
                # if it is a text append merge starting with it

                return self.__class__(
                    parts=(
                        *self.parts[:-1],
                        *_merge_texts(
                            text,
                            *chain.from_iterable(_extract_parts(part) for part in parts),
                        ),
                    )
                )

            case _:
                # otherwise just append merged items
                return self.__class__(
                    parts=(
                        *self.parts,
                        *_merge_texts(*chain.from_iterable(_extract_parts(part) for part in parts)),
                    )
                )

    def extended_by(
        self,
        *other: Self,
    ) -> Self:
        return self.appending(*chain.from_iterable(content.parts for content in other))

    def __bool__(self) -> bool:
        return bool(self.parts) and any(self.parts)

    def __str__(self) -> str:
        return self.to_str()


MultimodalContent.empty = MultimodalContent(parts=())
Multimodal = MultimodalContent | MultimodalContentConvertible


def _extract_parts(  # noqa: PLR0911
    element: Multimodal,
    /,
    meta: Meta | None = None,
) -> Sequence[MultimodalContentElement]:
    match element:
        case MultimodalContent() as content:
            if not content:
                return ()

            elif meta:
                return tuple(
                    _update_meta(
                        meta,
                        element=part,
                    )
                    for part in content.parts
                )

            else:
                return content.parts

        case str() as text:
            if not text:
                return ()

            else:
                return (
                    TextContent(
                        text=text,
                        meta=Meta.of(meta),
                    ),
                )

        case element:
            if not element:
                return ()

            elif meta:
                return (
                    _update_meta(
                        meta,
                        element=element,
                    ),
                )

            else:
                return (element,)


def _as_content(
    element: MultimodalContentConvertible,
    /,
) -> MultimodalContentElement:
    match element:
        case str() as text:
            return TextContent(text=text)

        case element:
            return element


def _is_media(
    element: MultimodalContentElement,
) -> bool:
    return isinstance(
        element,
        MediaContent,
    )


def _is_artifact(
    element: MultimodalContentElement,
) -> bool:
    return not isinstance(
        element,
        TextContent | MediaContent | MetaContent,
    )


def _as_string(
    element: MultimodalContentElement,
    /,
    *,
    include_data: bool,
) -> str:
    match element:
        case TextContent() as text:
            return text.text

        case MediaData() as media_data:
            return media_data.to_str(include_data=include_data)

        case MediaReference() as media_reference:
            return media_reference.to_str()

        case MetaContent() as meta:
            return (  # perhaps, we could use meta values within xml tag?
                (
                    f"<{meta.category}>"
                    f"{_as_string(meta.content, include_data=include_data)}"
                    f"</{meta.category}>"
                )
                if meta.content is not None
                else f"</{meta.category}>"
            )

        case DataModel() as model:
            return str(model)


def _merge_texts(
    *elements: MultimodalContentElement,
) -> Sequence[MultimodalContentElement]:
    if len(elements) <= 1:
        return elements

    result: list[MultimodalContentElement] = []
    last_text_element: TextContent | None = None
    for element in elements:
        match element:
            case TextContent() as text:
                # do not merge texts with different metadata
                if last_text := last_text_element:
                    if last_text.meta == text.meta:
                        last_text_element = TextContent(
                            text=last_text.text + text.text,
                            meta=text.meta,
                        )

                    else:
                        result.append(last_text)
                        last_text_element = text

                else:
                    last_text_element = text

            case other:
                if last_text := last_text_element:
                    result.append(last_text)
                    last_text_element = None

                result.append(other)

    if last_text := last_text_element:
        result.append(last_text)

    return result


def _update_meta(
    meta: Meta,
    /,
    element: MultimodalContentElement,
) -> MultimodalContentElement:
    match element:
        case (TextContent() | MediaData() | MediaReference() | MetaContent()) as content:
            if current_meta := content.meta:
                return content.updated(meta={**current_meta, **meta})

            else:
                return content.updated(meta=meta)

        case DataModel() as model:
            return model



from draive.prompts.state import Prompts
from draive.prompts.template import PromptAvailabilityCheck, PromptTemplate, prompt
from draive.prompts.types import (
    Prompt,
    PromptDeclaration,
    PromptDeclarationArgument,
    PromptException,
    PromptFetching,
    PromptListFetching,
    PromptMissing,
)

__all__ = (
    "Prompt",
    "PromptAvailabilityCheck",
    "PromptDeclaration",
    "PromptDeclarationArgument",
    "PromptException",
    "PromptFetching",
    "PromptListFetching",
    "PromptMissing",
    "PromptTemplate",
    "Prompts",
    "prompt",
)



from collections.abc import Callable, Iterable, Sequence
from typing import Any, Protocol, final, overload, runtime_checkable

from haiway import AttributePath, AttributeRequirement, State, ctx

from draive.multimodal import MediaContent, TextContent
from draive.parameters import DataModel

__all__ = (
    "VectorIndex",
    "VectorIndexing",
    "VectorSearching",
)


@runtime_checkable
class VectorIndexing(Protocol):
    async def __call__[Model: DataModel, Value: MediaContent | TextContent | str](
        self,
        model: type[Model],
        /,
        *,
        attribute: Callable[[Model], Value] | AttributePath[Model, Value] | Value,
        values: Iterable[Model],
        **extra: Any,
    ) -> None: ...


@runtime_checkable
class VectorSearching(Protocol):
    async def __call__[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        query: Sequence[float] | MediaContent | TextContent | str | None,
        score_threshold: float | None,
        requirements: AttributeRequirement[Model] | None,
        limit: int | None,
        **extra: Any,
    ) -> Iterable[Model]: ...


@runtime_checkable
class VectorDeleting(Protocol):
    async def __call__[Model: DataModel, Value: MediaContent | TextContent | str](
        self,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None,
        **extra: Any,
    ) -> None: ...


@final
class VectorIndex(State):
    @classmethod
    async def index[Model: DataModel, Value: MediaContent | TextContent | str](
        cls,
        model: type[Model],
        /,
        *,
        attribute: Callable[[Model], Value] | AttributePath[Model, Value] | Value,
        values: Iterable[Model],
        **extra: Any,
    ) -> None:
        return await ctx.state(cls).indexing(
            model,
            attribute=attribute,
            values=values,
            **extra,
        )

    @overload
    @classmethod
    async def search[Model: DataModel](
        cls,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int | None = None,
        **extra: Any,
    ) -> Iterable[Model]: ...

    @overload
    @classmethod
    async def search[Model: DataModel](
        cls,
        model: type[Model],
        /,
        *,
        query: Sequence[float] | MediaContent | TextContent | str,
        score_threshold: float | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int | None = None,
        **extra: Any,
    ) -> Iterable[Model]: ...

    @classmethod
    async def search[Model: DataModel](
        cls,
        model: type[Model],
        /,
        *,
        query: Sequence[float] | MediaContent | TextContent | str | None = None,
        score_threshold: float | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int | None = None,
        **extra: Any,
    ) -> Iterable[Model]:
        return await ctx.state(cls).searching(
            model,
            query=query,
            score_threshold=score_threshold,
            requirements=requirements,
            limit=limit,
            **extra,
        )

    @classmethod
    async def delete[Model: DataModel](
        cls,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        **extra: Any,
    ) -> None:
        return await ctx.state(cls).deleting(
            model,
            requirements=requirements,
            **extra,
        )

    indexing: VectorIndexing
    searching: VectorSearching
    deleting: VectorDeleting



from collections.abc import Callable, Coroutine
from typing import Protocol, final, overload

from haiway import ctx

from draive.commons import Meta, MetaValues
from draive.lmm import LMMContext
from draive.parameters import ParametrizedFunction
from draive.prompts.types import Prompt, PromptDeclaration, PromptDeclarationArgument

__all__ = (
    "PromptAvailabilityCheck",
    "PromptTemplate",
    "prompt",
)


class PromptAvailabilityCheck(Protocol):
    def __call__(self) -> bool: ...


@final
class PromptTemplate[**Args](ParametrizedFunction[Args, Coroutine[None, None, LMMContext]]):
    __slots__ = ("_check_availability", "declaration")

    def __init__(
        self,
        /,
        name: str,
        *,
        description: str | None = None,
        availability_check: PromptAvailabilityCheck | None,
        meta: Meta,
        function: Callable[Args, Coroutine[None, None, LMMContext]],
    ) -> None:
        super().__init__(function)

        self.declaration: PromptDeclaration
        object.__setattr__(
            self,
            "declaration",
            PromptDeclaration(
                name=name,
                description=description,
                arguments=[
                    PromptDeclarationArgument(
                        name=parameter.alias or parameter.name,
                        specification=parameter.specification,
                        required=parameter.required,
                    )
                    for parameter in self._parameters.values()
                ],
                meta=meta,
            ),
        )
        self._check_availability: PromptAvailabilityCheck
        object.__setattr__(
            self,
            "_check_availability",
            availability_check
            or (
                lambda meta: True  # available by default
            ),
        )

    @property
    def available(self) -> bool:
        try:
            return self._check_availability()

        except Exception:
            return False

    async def resolve(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Prompt:
        with ctx.scope(f"prompt:{self.declaration.name}"):
            try:
                result = Prompt(
                    name=self.declaration.name,
                    description=self.declaration.description,
                    content=await super().__call__(*args, **kwargs),  # pyright: ignore[reportCallIssue]
                )

                return result

            except BaseException as exc:
                ctx.log_error(
                    "Prompt resolving error",
                    exception=exc,
                )
                raise exc


class PromptTemplateWrapper(Protocol):
    def __call__[**Args](
        self,
        function: Callable[Args, Coroutine[None, None, LMMContext]],
    ) -> PromptTemplate[Args]: ...


@overload
def prompt[**Args](
    function: Callable[Args, Coroutine[None, None, LMMContext]],
    /,
) -> PromptTemplate[Args]: ...


@overload
def prompt[**Args](
    *,
    name: str | None = None,
    description: str | None = None,
    availability_check: PromptAvailabilityCheck | None = None,
    meta: Meta | MetaValues | None = None,
) -> PromptTemplateWrapper: ...


def prompt[**Args](
    function: Callable[Args, Coroutine[None, None, LMMContext]] | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
    availability_check: PromptAvailabilityCheck | None = None,
    meta: Meta | MetaValues | None = None,
) -> PromptTemplateWrapper | PromptTemplate[Args]:
    def wrap[**Arg](
        function: Callable[Arg, Coroutine[None, None, LMMContext]],
    ) -> PromptTemplate[Arg]:
        return PromptTemplate[Arg](
            name=name or function.__name__,
            description=description,
            availability_check=availability_check,
            meta=Meta.of(meta),
            function=function,
        )

    if function := function:
        return wrap(function=function)

    else:
        return wrap



from collections.abc import Iterable, Sequence
from itertools import chain
from typing import Any

from haiway import ctx

from draive.choice.types import ChoiceOption, SelectionException
from draive.instructions import Instruction
from draive.lmm import (
    LMM,
    LMMCompletion,
    LMMContextElement,
    LMMInput,
    LMMToolRequests,
    LMMToolResponses,
)
from draive.multimodal import Multimodal, MultimodalContent, MultimodalTagElement
from draive.tools import Toolbox

__all__ = ("default_choice_completion",)


async def default_choice_completion(  # noqa: C901
    *,
    instruction: Instruction | str,
    options: Sequence[ChoiceOption],
    input: Multimodal,  # noqa: A002
    toolbox: Toolbox,
    examples: Iterable[tuple[Multimodal, ChoiceOption]] | None,
    **extra: Any,
) -> ChoiceOption:
    with ctx.scope("choice_completion"):
        assert options, "Choice options cannot be empty"  # nosec: B101
        assert all(  # nosec: B101
            example[1] in options for example in examples or []
        ), "Choice options examples have to be included in available options"

        if len(options) == 1:
            return options[0]  # if there is ony one option it has to be the result

        options_map: dict[str, ChoiceOption] = {option.identifier: option for option in options}
        assert len(options_map) == len(options), "Choice options cannot have duplicate identifiers"  # nosec: B101

        extended_instruction: Instruction
        match instruction:
            case str() as instruction_string:
                extended_instruction = Instruction.of(
                    f"{instruction_string}\n\n{INSTRUCTION_EXTENSION}"
                )

            case instruction:
                extended_instruction = instruction.extended(
                    INSTRUCTION_EXTENSION,
                    joiner="\n\n",
                )

        formatted_options: MultimodalContent = MultimodalContent.of(
            *(_format_option(option) for option in options)
        )

        context: list[LMMContextElement] = [
            *chain.from_iterable(
                _format_example(
                    example,
                    options=formatted_options,
                )
                for example in examples or []
            ),
            _format_input(
                input,
                options=formatted_options,
            ),
        ]

        recursion_level: int = 0
        while recursion_level <= toolbox.repeated_calls_limit:
            match await LMM.completion(
                instruction=extended_instruction,
                context=context,
                tool_selection=toolbox.tool_selection(repetition_level=recursion_level),
                tools=toolbox.available_tools(),
                output="text",
                **extra,
            ):
                case LMMCompletion() as completion:
                    ctx.log_debug("Received choice results")
                    if selection := MultimodalTagElement.parse_first(
                        "CHOICE",
                        content=completion.content,
                    ):
                        if option := options_map.get(selection.content.to_str()):
                            return option

                    raise SelectionException("Invalid or missing selection")

                case LMMToolRequests() as tool_requests:
                    ctx.log_debug("Received choice tool calls")
                    tool_responses: LMMToolResponses = await toolbox.respond_all(tool_requests)

                    if direct_content := [
                        response.content
                        for response in tool_responses.responses
                        if response.handling == "direct_result"
                    ]:
                        if selection := MultimodalTagElement.parse_first(
                            "CHOICE",
                            content=MultimodalContent.of(*direct_content),
                        ):
                            if option := options_map.get(selection.content.to_str()):
                                return option

                        raise SelectionException("Invalid or missing selection")

                    else:
                        context.extend(
                            [
                                tool_requests,
                                tool_responses,
                            ]
                        )

            recursion_level += 1  # continue with next recursion level

        raise RuntimeError("LMM exceeded limit of recursive calls")


def _format_option(
    option: ChoiceOption,
    /,
) -> MultimodalContent:
    return MultimodalContent.of(
        f"<OPTION identifier='{option.identifier}'>",
        option.content,
        "</OPTION>",
    )


def _format_input(
    input: Multimodal,  # noqa: A002
    /,
    options: MultimodalContent,
) -> LMMInput:
    return LMMInput.of(
        MultimodalContent.of(
            "<INPUT>",
            input,
            "</INPUT>/n/n<OPTIONS>",
            options,
            "</OPTIONS>",
        )
    )


def _format_example(
    example: tuple[Multimodal, ChoiceOption],
    /,
    options: MultimodalContent,
) -> tuple[LMMInput, LMMCompletion]:
    return (
        _format_input(example[0], options=options),
        LMMCompletion.of(f"```\n{example[1].identifier}\n```"),
    )


INSTRUCTION_EXTENSION: str = """\
<FORMAT>
Ensure the result is following the strict format:
Place only the identifier of the chosen option inside a <CHOICE> tag, as shown here:

<CHOICE>option_identifier</CHOICE>

</FORMAT>
"""



from haiway import getenv_str
from mistralai import Mistral as MistralClient

__all__ = ("MistralAPI",)


class MistralAPI:
    __slots__ = (
        "_api_key",
        "_client",
        "_server_url",
        "_timeout_ms",
    )

    def __init__(
        self,
        server_url: str | None = None,
        api_key: str | None = None,
        timeout: float = 60.0,
    ) -> None:
        self._server_url: str = server_url or getenv_str(
            "MISTRAL_SERVER_URL",
            default="https://api.mistral.ai",
        )
        self._api_key: str | None = api_key or getenv_str("MISTRAL_API_KEY")
        self._timeout_ms: int = int(timeout * 1000)
        self._client = self._prepare_client()

    def _prepare_client(self) -> MistralClient:
        return MistralClient(
            api_key=self._api_key,
            server_url=self._server_url,
            timeout_ms=self._timeout_ms,
        )

    async def _initialize_client(self) -> None:
        if self._client.sdk_configuration.async_client is not None:
            await self._client.sdk_configuration.async_client.aclose()
        self._client = self._prepare_client()
        await self._client.__aenter__()

    async def _deinitialize_client(self) -> None:
        await self._client.__aexit__(
            None,
            None,
            None,
        )



from haiway import getenv_str
from openai import AsyncAzureOpenAI, AsyncOpenAI

__all__ = ("OpenAIAPI",)


class OpenAIAPI:
    __slots__ = (
        "_api_key",
        "_azure_api_endpoint",
        "_azure_api_version",
        "_azure_deployment",
        "_base_url",
        "_client",
        "_organization",
    )

    def __init__(
        self,
        base_url: str | None = None,
        api_key: str | None = None,
        organization: str | None = None,
        azure_api_endpoint: str | None = None,
        azure_api_version: str | None = None,
        azure_deployment: str | None = None,
    ) -> None:
        self._base_url: str | None = base_url or getenv_str("OPENAI_BASE_URL")
        self._api_key: str = api_key or getenv_str(
            "AZURE_OPENAI_API_KEY",
            default=getenv_str(
                "OPENAI_API_KEY",
                default="",
            ),
        )
        self._organization: str | None = organization or getenv_str("OPENAI_ORGANIZATION")
        self._azure_api_endpoint: str | None = azure_api_endpoint or getenv_str(
            "AZURE_OPENAI_API_BASE"
        )
        self._azure_deployment: str | None = azure_deployment or getenv_str(
            "AZURE_OPENAI_DEPLOYMENT_NAME"
        )
        self._azure_api_version: str | None = azure_api_version or getenv_str(
            "AZURE_OPENAI_API_VERSION"
        )
        self._client: AsyncOpenAI = self._prepare_client()

    def _prepare_client(self) -> AsyncOpenAI:
        if self._azure_api_endpoint:
            return AsyncAzureOpenAI(
                api_key=self._api_key,
                azure_endpoint=self._azure_api_endpoint,
                azure_deployment=self._azure_deployment,
                api_version=self._azure_api_version,
                organization=self._organization,
            )

        else:  # otherwise try using OpenAI
            return AsyncOpenAI(
                base_url=self._base_url,
                api_key=self._api_key,
                organization=self._organization,
            )

    async def _initialize_client(self) -> None:
        await self._client.close()
        self._client = self._prepare_client()
        await self._client.__aenter__()

    async def _deinitialize_client(self) -> None:
        await self._client.__aexit__(
            None,
            None,
            None,
        )



from typing import Any

from draive.evaluation import EvaluationScore
from draive.parameters import DataModel, Field, ParameterValidationContext

__all__ = ("CommonScoreModel",)


def _score_validator(
    value: Any,
    context: ParameterValidationContext,
) -> float:
    match value:
        case float() as float_value:
            return float_value

        case int() as int_value:
            return float(int_value)

        case str() as str_value:
            try:
                return float(str_value)

            except Exception as exc:
                raise TypeError(f"Expected 'float', received '{type(value).__name__}'") from exc

        case _:
            raise TypeError(f"Expected 'float', received '{type(value).__name__}'")


class CommonScoreModel(DataModel):
    comment: str | None = Field(
        description="Explanation of the score",
        default=None,
    )
    score: float = Field(
        description="Decimal score value",
        validator=_score_validator,
    )

    def normalized(
        self,
        divider: float | None = None,
    ) -> EvaluationScore:
        return EvaluationScore(
            value=self.score / divider if divider else self.score,
            comment=self.comment,
        )



from collections.abc import AsyncIterator, Iterable, Mapping, Sequence
from typing import (
    Any,
    Literal,
    Protocol,
    Self,
    TypedDict,
    overload,
    runtime_checkable,
)

from haiway import Default

from draive.commons import META_EMPTY, Meta, MetaValues
from draive.instructions import Instruction
from draive.multimodal import Multimodal, MultimodalContent
from draive.parameters import DataModel, ParametersSpecification

__all__ = (
    "LMMCompleting",
    "LMMCompletion",
    "LMMContext",
    "LMMContextElement",
    "LMMException",
    "LMMInput",
    "LMMOutput",
    "LMMSessionEvent",
    "LMMSessionOutput",
    "LMMSessionOutputSelection",
    "LMMStreamChunk",
    "LMMStreamInput",
    "LMMStreamOutput",
    "LMMToolError",
    "LMMToolFunctionSpecification",
    "LMMToolRequest",
    "LMMToolRequests",
    "LMMToolResponse",
    "LMMToolResponseHandling",
    "LMMToolResponses",
    "LMMToolSelection",
    "LMMToolSpecification",
)


class LMMToolFunctionSpecification(TypedDict):
    name: str
    description: str | None
    parameters: ParametersSpecification | None


type LMMToolSpecification = LMMToolFunctionSpecification
type LMMOutputSelection = (
    Sequence[Literal["text", "image", "audio", "video"]]
    | Literal["auto", "text", "json", "image", "audio", "video"]
    | type[DataModel]
)
type LMMToolSelection = Literal["auto", "required", "none"] | LMMToolSpecification


class LMMException(Exception):
    pass


class LMMToolError(LMMException):
    def __init__(
        self,
        *args: object,
        content: MultimodalContent,
    ) -> None:
        super().__init__(*args)
        self.content: MultimodalContent = content


class LMMInput(DataModel):
    @classmethod
    def of(
        cls,
        content: Multimodal,
        /,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            content=MultimodalContent.of(content),
            meta=Meta.of(meta),
        )

    type: Literal["input"] = "input"
    content: MultimodalContent
    meta: Meta = META_EMPTY

    def __bool__(self) -> bool:
        return bool(self.content)


class LMMCompletion(DataModel):
    @classmethod
    def of(
        cls,
        content: Multimodal,
        /,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            content=MultimodalContent.of(content),
            meta=Meta.of(meta),
        )

    type: Literal["completion"] = "completion"
    content: MultimodalContent
    meta: Meta = META_EMPTY

    def __bool__(self) -> bool:
        return bool(self.content)


LMMToolResponseHandling = Literal["error", "result", "direct_result"]


class LMMToolResponse(DataModel):
    type: Literal["tool_response"] = "tool_response"
    identifier: str
    tool: str
    content: MultimodalContent
    handling: LMMToolResponseHandling


class LMMToolResponses(DataModel):
    @classmethod
    def of(
        cls,
        responses: Sequence[LMMToolResponse],
        /,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            responses=responses,
            meta=Meta.of(meta),
        )

    type: Literal["tool_responses"] = "tool_responses"
    responses: Sequence[LMMToolResponse]
    meta: Meta = META_EMPTY


class LMMToolRequest(DataModel):
    @classmethod
    def of(
        cls,
        identifier: str,
        /,
        tool: str,
        arguments: Mapping[str, Any] | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            identifier=identifier,
            tool=tool,
            arguments=arguments if arguments is not None else {},
            meta=Meta.of(meta),
        )

    type: Literal["tool_request"] = "tool_request"
    identifier: str
    tool: str
    arguments: Mapping[str, Any] = Default(factory=dict)
    meta: Meta = META_EMPTY


class LMMToolRequests(DataModel):
    @classmethod
    def of(
        cls,
        requests: Sequence[LMMToolRequest],
        /,
        content: MultimodalContent | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            requests=requests,
            content=content,
            meta=Meta.of(meta),
        )

    type: Literal["tool_requests"] = "tool_requests"
    requests: Sequence[LMMToolRequest]
    content: MultimodalContent | None = None
    meta: Meta = META_EMPTY


LMMContextElement = LMMInput | LMMCompletion | LMMToolRequests | LMMToolResponses
LMMContext = Sequence[LMMContextElement]
LMMOutput = LMMCompletion | LMMToolRequests


class LMMStreamChunk(DataModel):
    @classmethod
    def of(
        cls,
        content: Multimodal,
        /,
        eod: bool = False,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            content=MultimodalContent.of(content),
            eod=eod,
            meta=Meta.of(meta),
        )

    type: Literal["stream_chunk"] = "stream_chunk"
    content: MultimodalContent
    eod: bool
    meta: Meta = META_EMPTY

    def __bool__(self) -> bool:
        return bool(self.content)


LMMStreamInput = LMMStreamChunk | LMMToolResponse
LMMStreamOutput = LMMStreamChunk | LMMToolRequest


class LMMSessionEvent(DataModel):
    @classmethod
    def of(
        cls,
        category: Literal["completed", "interrupted"] | str,
        /,
        *,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            category=category,
            meta=Meta.of(meta),
        )

    type: Literal["session_event"] = "session_event"
    category: Literal["completed", "interrupted"] | str
    meta: Meta


LMMSessionOutput = LMMStreamChunk | LMMToolRequest | LMMSessionEvent


@runtime_checkable
class LMMCompleting(Protocol):
    @overload
    async def __call__(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        stream: Literal[False] = False,
        **extra: Any,
    ) -> LMMOutput: ...

    @overload
    async def __call__(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        stream: Literal[True],
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput]: ...

    async def __call__(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        stream: bool = False,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput] | LMMOutput: ...


type LMMSessionOutputSelection = (
    Sequence[Literal["text", "audio"]] | Literal["auto", "text", "audio"]
)


@runtime_checkable
class LMMSessionPreparing(Protocol):
    async def __call__(
        self,
        *,
        instruction: Instruction | None,
        initial_context: LMMContext | None,
        input_stream: AsyncIterator[LMMStreamInput],
        output: LMMSessionOutputSelection,
        tools: Sequence[LMMToolSpecification],
        tool_selection: LMMToolSelection,
        **extra: Any,
    ) -> AsyncIterator[LMMSessionOutput]: ...



from collections.abc import AsyncIterator, Iterable, Sequence
from typing import Any, Literal, final, overload

from haiway import State, ctx

from draive.instructions import Instruction
from draive.lmm.types import (
    LMMCompleting,
    LMMContext,
    LMMOutput,
    LMMOutputSelection,
    LMMSessionOutput,
    LMMSessionOutputSelection,
    LMMSessionPreparing,
    LMMStreamInput,
    LMMStreamOutput,
    LMMToolSelection,
    LMMToolSpecification,
)

__all__ = (
    "LMM",
    "LMMSession",
)


@final
class LMM(State):
    @overload
    @classmethod
    async def completion(
        cls,
        *,
        instruction: Instruction | str | None = None,
        context: LMMContext,
        tool_selection: LMMToolSelection = "auto",
        tools: Iterable[LMMToolSpecification] | None = None,
        output: LMMOutputSelection = "auto",
        stream: Literal[False] = False,
        **extra: Any,
    ) -> LMMOutput: ...

    @overload
    @classmethod
    async def completion(
        cls,
        *,
        instruction: Instruction | str | None = None,
        context: LMMContext,
        tool_selection: LMMToolSelection = "auto",
        tools: Iterable[LMMToolSpecification] | None = None,
        output: LMMOutputSelection = "auto",
        stream: Literal[True],
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput]: ...

    @classmethod
    async def completion(
        cls,
        *,
        instruction: Instruction | str | None = None,
        context: LMMContext,
        tool_selection: LMMToolSelection = "auto",
        tools: Iterable[LMMToolSpecification] | None = None,
        output: LMMOutputSelection = "auto",
        stream: bool = False,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput] | LMMOutput:
        if stream:
            return await ctx.state(cls).completing(
                instruction=Instruction.of(instruction),
                context=context,
                tool_selection=tool_selection,
                tools=tools,
                output=output,
                stream=True,
                **extra,
            )

        else:
            return await ctx.state(cls).completing(
                instruction=Instruction.of(instruction),
                context=context,
                tool_selection=tool_selection,
                tools=tools,
                output=output,
                **extra,
            )

    completing: LMMCompleting


@final
class LMMSession(State):
    @classmethod
    async def prepare(
        cls,
        *,
        instruction: Instruction | None = None,
        initial_context: LMMContext | None = None,
        input_stream: AsyncIterator[LMMStreamInput],
        output: LMMSessionOutputSelection = "auto",
        tools: Sequence[LMMToolSpecification] | None = None,
        tool_selection: LMMToolSelection = "auto",
        **extra: Any,
    ) -> AsyncIterator[LMMSessionOutput]:
        return await ctx.state(cls).preparing(
            instruction=instruction,
            initial_context=initial_context,
            input_stream=input_stream,
            output=output,
            tools=tools if tools is not None else (),
            tool_selection=tool_selection,
            **extra,
        )

    preparing: LMMSessionPreparing



from draive.evaluators.coherence import coherence_evaluator
from draive.evaluators.conciseness import conciseness_evaluator
from draive.evaluators.consistency import consistency_evaluator
from draive.evaluators.coverage import coverage_evaluator
from draive.evaluators.fluency import fluency_evaluator
from draive.evaluators.groundedness import groundedness_evaluator
from draive.evaluators.keywords import keywords_evaluator
from draive.evaluators.readability import readability_evaluator
from draive.evaluators.relevance import relevance_evaluator
from draive.evaluators.similarity import (
    image_vector_similarity_evaluator,
    similarity_evaluator,
    text_vector_similarity_evaluator,
)
from draive.evaluators.truthfulness import truthfulness_evaluator

__all__ = (
    "coherence_evaluator",
    "conciseness_evaluator",
    "consistency_evaluator",
    "coverage_evaluator",
    "fluency_evaluator",
    "groundedness_evaluator",
    "image_vector_similarity_evaluator",
    "keywords_evaluator",
    "readability_evaluator",
    "relevance_evaluator",
    "similarity_evaluator",
    "text_vector_similarity_evaluator",
    "truthfulness_evaluator",
)



from draive.evaluation.evaluator import (
    Evaluator,
    EvaluatorDefinition,
    EvaluatorResult,
    PreparedEvaluator,
    evaluator,
)
from draive.evaluation.generator import generate_case_parameters
from draive.evaluation.scenario import (
    EvaluationScenarioResult,
    PreparedScenarioEvaluator,
    ScenarioEvaluator,
    ScenarioEvaluatorDefinition,
    ScenarioEvaluatorResult,
    evaluation_scenario,
)
from draive.evaluation.score import EvaluationScore
from draive.evaluation.suite import (
    EvaluationCaseResult,
    EvaluationSuite,
    EvaluationSuiteCase,
    EvaluationSuiteDefinition,
    EvaluationSuiteStorage,
    SuiteEvaluatorCaseResult,
    SuiteEvaluatorResult,
    evaluation_suite,
)
from draive.evaluation.value import EvaluationScoreValue

__all__ = (
    "EvaluationCaseResult",
    "EvaluationScenarioResult",
    "EvaluationScore",
    "EvaluationScoreValue",
    "EvaluationSuite",
    "EvaluationSuiteCase",
    "EvaluationSuiteDefinition",
    "EvaluationSuiteStorage",
    "Evaluator",
    "EvaluatorDefinition",
    "EvaluatorResult",
    "PreparedEvaluator",
    "PreparedScenarioEvaluator",
    "ScenarioEvaluator",
    "ScenarioEvaluatorDefinition",
    "ScenarioEvaluatorResult",
    "SuiteEvaluatorCaseResult",
    "SuiteEvaluatorResult",
    "evaluation_scenario",
    "evaluation_suite",
    "evaluator",
    "generate_case_parameters",
)



from collections.abc import Iterable
from typing import Any, Literal

from haiway import State, ctx

from draive.generation.model.default import generate_model
from draive.generation.model.types import ModelGenerating, ModelGeneratorDecoder
from draive.instructions import Instruction
from draive.multimodal import Multimodal
from draive.parameters import DataModel
from draive.prompts import Prompt
from draive.tools import Tool, Toolbox

__all__ = ("ModelGeneration",)


class ModelGeneration(State):
    @classmethod
    async def generate[Generated: DataModel](
        cls,
        generated: type[Generated],
        /,
        *,
        instruction: Instruction | str,
        input: Prompt | Multimodal,  # noqa: A002
        schema_injection: Literal["auto", "full", "simplified", "skip"] = "auto",
        tools: Toolbox | Iterable[Tool] | None = None,
        examples: Iterable[tuple[Multimodal, Generated]] | None = None,
        decoder: ModelGeneratorDecoder | None = None,
        **extra: Any,
    ) -> Generated:
        return await ctx.state(cls).generating(
            generated,
            instruction=instruction,
            input=input,
            schema_injection=schema_injection,
            tools=tools,
            examples=examples,
            decoder=decoder,
            **extra,
        )

    generating: ModelGenerating = generate_model



from draive.realtime.state import Realtime
from draive.realtime.types import RealtimeOutputSelection, RealtimeProcessing

__all__ = (
    "Realtime",
    "RealtimeOutputSelection",
    "RealtimeProcessing",
)



from typing import Any, Protocol, runtime_checkable

from draive.commons import Meta, MetaValues
from draive.multimodal import MultimodalContent

__all__ = (
    "GuardrailsQualityException",
    "GuardrailsQualityVerifying",
)


class GuardrailsQualityException(Exception):
    def __init__(
        self,
        *args: object,
        reason: str,
        content: MultimodalContent,
        meta: Meta | MetaValues | None = None,
    ) -> None:
        super().__init__(*args)
        self.reason: str = reason
        self.content: MultimodalContent = content
        self.meta: Meta = Meta.of(meta)


@runtime_checkable
class GuardrailsQualityVerifying(Protocol):
    async def __call__(
        self,
        content: MultimodalContent,
        /,
        **extra: Any,
    ) -> None: ...



import json
from collections.abc import Mapping, Sequence
from copy import copy, deepcopy
from datetime import UTC, datetime
from typing import Any, Literal, NotRequired, Required, TypedDict
from uuid import UUID, uuid4

from haiway import Default
from pytest import raises

from draive import (
    MISSING,
    ConversationMessage,
    DataModel,
    Field,
    Missing,
    MultimodalContent,
    ParameterValidationError,
)
from draive.multimodal import MediaData, MediaReference


def invalid(value: str) -> None:
    if value != "valid":
        raise ValueError()


class ExampleNestedModel(DataModel):
    nested_alias: str = Field(aliased="string", default="default")
    more: Sequence[int] | None = None


class ExampleModel(DataModel):
    string: str = "default"
    number: int = Field(aliased="alias", description="description", default=1)
    none_default: int | None = Field(default=None)
    value_default: int = Field(default=9)
    invalid: str = Field(verifier=invalid, default="valid")
    nested: ExampleNestedModel = Field(aliased="answer", default=ExampleNestedModel())
    full: Literal["A", "B"] | Sequence[int] | str | bool | None = Field(
        aliased="all",
        description="complex",
        default="",
    )
    default: UUID = Default(factory=uuid4)


# TODO: prepare extensive tests
def test_validated_passes_with_valid_values() -> None:
    values_dict: dict[str, Any] = {
        "string": "value",
        "alias": 42,
        "none_default": 1,
        "value_default": 8,
        "invalid": "valid",
        "answer": {"string": "value", "more": None},
        "all": True,
        "default": "12345678123456781234567812345678",
    }
    assert json.loads(ExampleModel(**values_dict).to_json()) == values_dict


def test_validated_passes_with_default_values() -> None:
    ExampleModel()


def test_default_factory_produces_values() -> None:
    assert ExampleModel().default != ExampleModel().default


class DatetimeModel(DataModel):
    value: datetime


datetime_model_instance: DatetimeModel = DatetimeModel(
    value=datetime.fromtimestamp(0, UTC),
)
datetime_model_json: str = '{"value": "1970-01-01T00:00:00+00:00"}'


def test_datetime_encoding() -> None:
    assert datetime_model_instance.to_json() == datetime_model_json


def test_datetime_decoding() -> None:
    assert DatetimeModel.from_json(datetime_model_json) == datetime_model_instance


class UUIDModel(DataModel):
    value: UUID


uuid_model_instance: UUIDModel = UUIDModel(
    value=UUID(hex="0cf728c0369348e78552e8d86d35e8b0"),
)
uuid_model_json: str = '{"value": "0cf728c0369348e78552e8d86d35e8b0"}'


def test_uuid_encoding() -> None:
    assert uuid_model_instance.to_json() == uuid_model_json


def test_uuid_decoding() -> None:
    assert UUIDModel.from_json(uuid_model_json) == uuid_model_instance


class MissingModel(DataModel):
    value: Missing = MISSING


missing_model_instance: MissingModel = MissingModel(
    value=MISSING,
)
missing_model_json: str = "{}"


def test_missing_encoding() -> None:
    assert missing_model_instance.to_json(indent=4) == missing_model_json


def test_missing_decoding() -> None:
    assert MissingModel.from_json(missing_model_json) == missing_model_instance


class ExampleTypedDictNested(DataModel):
    value: int


class ExampleTypedDict(TypedDict, total=False):
    dict_int: Required[int]
    dict_str: NotRequired[str | None]
    dict_nested: Required[ExampleTypedDictNested]


class ExampleTotalTypedDict(TypedDict, total=True):
    total: bool


class TypedDictModel(DataModel):
    value: ExampleTypedDict
    total_value: ExampleTotalTypedDict


typed_dict_model_instance: TypedDictModel = TypedDictModel(
    value={
        "dict_int": 42,
        "dict_str": "answer",
        "dict_nested": ExampleTypedDictNested(
            value=21,
        ),
    },
    total_value={"total": True},
)
typed_dict_model_json: str = """\
{
    "value": {
        "dict_int": 42,
        "dict_str": "answer",
        "dict_nested": {
            "value": 21
        }
    },
    "total_value": {
        "total": true
    }
}\
"""
typed_dict_not_required_model_instance: TypedDictModel = TypedDictModel(
    value={
        "dict_int": 42,
        "dict_nested": ExampleTypedDictNested(
            value=21,
        ),
    },
    total_value={"total": True},
)
typed_dict_not_required_model_json: str = """\
{
    "value": {
        "dict_int": 42,
        "dict_nested": {
            "value": 21
        }
    },
    "total_value": {
        "total": true
    }
}\
"""


def test_typed_dict_encoding() -> None:
    assert typed_dict_model_instance.to_json(indent=4) == typed_dict_model_json
    assert (
        typed_dict_not_required_model_instance.to_json(indent=4)
        == typed_dict_not_required_model_json
    )


def test_typed_dict_decoding() -> None:
    assert TypedDictModel.from_json(typed_dict_model_json) == typed_dict_model_instance
    assert (
        TypedDictModel.from_json(typed_dict_not_required_model_json)
        == typed_dict_not_required_model_instance
    )


class BasicsModel(DataModel):
    string: str
    string_list: Sequence[str]
    integer: int
    integer_or_float_list: Sequence[int | float]
    floating: float
    floating_dict: Mapping[str, float]
    optional: str | None
    none: None


basic_model_instance: BasicsModel = BasicsModel(
    string="test",
    string_list=["basic", "list"],
    integer=42,
    integer_or_float_list=[12, 3.14, 7],
    floating=9.99,
    floating_dict={"a": 65, "b": 66.0, "c": 67.5},
    optional="some",
    none=None,
)
basic_model_json: str = """\
{
    "string": "test",
    "string_list": [
        "basic",
        "list"
    ],
    "integer": 42,
    "integer_or_float_list": [
        12,
        3.14,
        7
    ],
    "floating": 9.99,
    "floating_dict": {
        "a": 65.0,
        "b": 66.0,
        "c": 67.5
    },
    "optional": "some",
    "none": null
}\
"""


def test_basic_encoding() -> None:
    assert basic_model_instance.to_json(indent=4) == basic_model_json


def test_basic_decoding() -> None:
    assert BasicsModel.from_json(basic_model_json) == basic_model_instance


basic_conversation_message_instance: ConversationMessage = ConversationMessage(
    identifier=UUID(hex="f6da0a47556744cdb8334d263020907f"),
    role="model",
    created=datetime.fromisoformat("2025-06-03T18:15:58.985599"),
    content=MultimodalContent.of("string"),
)
basic_conversation_message_json: str = """\
{
    "type": "message",
    "identifier": "f6da0a47556744cdb8334d263020907f",
    "role": "model",
    "created": "2025-06-03T18:15:58.985599",
    "content": {
        "parts": [
            {
                "text": "string",
                "meta": {}
            }
        ]
    },
    "meta": {}
}\
"""

media_url_conversation_message_instance: ConversationMessage = ConversationMessage(
    identifier=UUID(hex="f6da0a47556744cdb8334d263020907f"),
    role="model",
    created=datetime.fromisoformat("2025-06-03T18:15:58.985599"),
    content=MultimodalContent.of(MediaReference.of("https://miquido.com/image", media="image/png")),
)
media_url_conversation_message_json: str = """\
{
    "type": "message",
    "identifier": "f6da0a47556744cdb8334d263020907f",
    "role": "model",
    "created": "2025-06-03T18:15:58.985599",
    "content": {
        "parts": [
            {
                "media": "image/png",
                "uri": "https://miquido.com/image",
                "meta": {}
            }
        ]
    },
    "meta": {}
}\
"""

media_data_conversation_message_instance: ConversationMessage = ConversationMessage(
    identifier=UUID(hex="f6da0a47556744cdb8334d263020907f"),
    role="model",
    created=datetime.fromisoformat("2025-06-03T18:15:58.985599"),
    content=MultimodalContent.of(MediaData.of(b"image_data", media="image/png")),
)
media_data_conversation_message_json: str = """\
{
    "type": "message",
    "identifier": "f6da0a47556744cdb8334d263020907f",
    "role": "model",
    "created": "2025-06-03T18:15:58.985599",
    "content": {
        "parts": [
            {
                "media": "image/png",
                "data": "aW1hZ2VfZGF0YQ==",
                "meta": {}
            }
        ]
    },
    "meta": {}
}\
"""


def test_llm_message_decoding() -> None:
    assert (
        ConversationMessage.from_json(basic_conversation_message_json)
        == basic_conversation_message_instance
    )
    assert (
        ConversationMessage.from_json(media_url_conversation_message_json)
        == media_url_conversation_message_instance
    )
    assert (
        ConversationMessage.from_json(media_data_conversation_message_json)
        == media_data_conversation_message_instance
    )


def test_llm_message_encoding() -> None:
    assert basic_conversation_message_instance.to_json(indent=4) == basic_conversation_message_json
    assert (
        media_url_conversation_message_instance.to_json(indent=4)
        == media_url_conversation_message_json
    )
    assert (
        media_data_conversation_message_instance.to_json(indent=4)
        == media_data_conversation_message_json
    )


class AnyModelNested(DataModel):
    answer: int


class AnyModel(DataModel):
    any_model: DataModel


any_model_instance: AnyModel = AnyModel(
    any_model=AnyModelNested(
        answer=42,
    ),
)
any_model_json: str = """\
{
    "any_model": {
        "answer": 42
    }
}\
"""


def test_any_encoding() -> None:
    assert any_model_instance.to_json(indent=4) == any_model_json


def test_any_dict_decoding() -> None:
    assert DataModel.from_json(any_model_json).to_mapping() == any_model_instance.to_mapping()


def test_generic_subtypes_validation() -> None:
    class NestedGeneric[T](DataModel):
        value: T

    class Generic[T](DataModel):
        nested: NestedGeneric[T]

    class Container(DataModel):
        generic: Generic[str]

    assert isinstance(Generic[str](nested=NestedGeneric[str](value="ok")), Generic)
    assert isinstance(Generic(nested=NestedGeneric[str](value="ok")), Generic)
    assert isinstance(Generic(nested=NestedGeneric(value="ok")), Generic)

    assert isinstance(Generic[str](nested=NestedGeneric[str](value="ok")), Generic[Any])
    assert isinstance(Generic(nested=NestedGeneric[str](value="ok")), Generic[Any])
    assert isinstance(Generic[str](nested=NestedGeneric(value="ok")), Generic[Any])
    assert isinstance(Generic(nested=NestedGeneric(value="ok")), Generic[Any])

    assert isinstance(Generic[str](nested=NestedGeneric[str](value="ok")), Generic[str])
    assert isinstance(Generic(nested=NestedGeneric[str](value="ok")), Generic[str])
    assert isinstance(Generic[str](nested=NestedGeneric(value="ok")), Generic[str])
    assert isinstance(Generic(nested=NestedGeneric(value="ok")), Generic[str])

    with raises(ParameterValidationError):
        _ = Generic[int](nested=NestedGeneric[str](value="ok"))

    with raises(ParameterValidationError):
        _ = Container(generic=Generic(nested=NestedGeneric(value=42)))

    with raises(ParameterValidationError):
        _ = Container(generic=Generic[int](nested=NestedGeneric[str](value="ok")))

    # not raises
    _ = Container(generic=Generic(nested=NestedGeneric(value="ok")))


def test_copying_leaves_same_object() -> None:
    class Nested(DataModel):
        string: str

    class Copied(DataModel):
        string: str
        nested: Nested

    origin = Copied(string="42", nested=Nested(string="answer"))
    assert copy(origin) is origin
    assert deepcopy(origin) is origin



from typing import Literal

from draive.configuration import Config

__all__ = (
    "CohereImageEmbeddingConfig",
    "CohereTextEmbeddingConfig",
)


class CohereTextEmbeddingConfig(Config):
    model: str
    purpose: Literal[
        "search_query",
        "search_document",
        "clustering",
        "classification",
    ]
    batch_size: int = 128


class CohereImageEmbeddingConfig(Config):
    model: str
    batch_size: int = 16



from collections.abc import Callable, Sequence
from typing import Any, Protocol, runtime_checkable

from haiway import State

from draive.parameters import DataModel

__all__ = (
    "Embedded",
    "ValueEmbedding",
)


class Embedded[Value: DataModel | State | str | bytes](State):
    value: Value
    vector: Sequence[float]


@runtime_checkable
class ValueEmbedding[Value: DataModel | State | str | bytes, Data: str | bytes](Protocol):
    # @overload
    # async def __call__(
    #     self,
    #     values: Sequence[Data],
    #     /,
    #     **extra: Any,
    # ) -> Sequence[Embedded[Data]]: ...

    # @overload
    # async def __call__(
    #     self,
    #     values: Sequence[Value],
    #     /,
    #     attribute: Callable[[Value], Data],
    #     **extra: Any,
    # ) -> Sequence[Embedded[Value]]: ...

    async def __call__(
        self,
        values: Sequence[Value] | Sequence[Data],
        /,
        attribute: Callable[[Value], Data] | None = None,
        **extra: Any,
    ) -> Sequence[Embedded[Value]] | Sequence[Embedded[Data]]: ...



import json
import typing
from collections.abc import Callable, Iterator, Mapping
from copy import deepcopy
from dataclasses import fields as dataclass_fields
from dataclasses import is_dataclass
from datetime import date, datetime, time
from types import EllipsisType, GenericAlias
from typing import (
    Any,
    ClassVar,
    Generic,
    Self,
    TypeVar,
    cast,
    dataclass_transform,
    final,
    overload,
)
from uuid import UUID
from weakref import WeakValueDictionary

from haiway import MISSING, AttributePath, DefaultValue, Missing, not_missing
from haiway.state import AttributeAnnotation, attribute_annotations

from draive.parameters.parameter import Parameter
from draive.parameters.schema import json_schema, simplified_schema
from draive.parameters.specification import (
    ParameterSpecification,
    ParametersSpecification,
)
from draive.parameters.types import (
    ParameterConversion,
    ParameterValidation,
    ParameterValidationContext,
    ParameterVerification,
)

__all__ = (
    "DataModel",
    "Field",
)


@overload
def Field[Value](
    *,
    aliased: str | None = None,
    description: str | Missing = MISSING,
    default: Value | Missing = MISSING,
    validator: ParameterValidation[Value] | Missing = MISSING,
    converter: ParameterConversion[Value] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


@overload
def Field[Value](
    *,
    aliased: str | None = None,
    description: str | Missing = MISSING,
    default_factory: Callable[[], Value] | Missing = MISSING,
    validator: ParameterValidation[Value] | Missing = MISSING,
    converter: ParameterConversion[Value] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


@overload
def Field[Value](
    *,
    aliased: str | None = None,
    description: str | Missing = MISSING,
    default: Value | Missing = MISSING,
    verifier: ParameterVerification[Value] | Missing = MISSING,
    converter: ParameterConversion[Value] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


@overload
def Field[Value](
    *,
    aliased: str | None = None,
    description: str | Missing = MISSING,
    default_factory: Callable[[], Value] | Missing = MISSING,
    verifier: ParameterVerification[Value] | Missing = MISSING,
    converter: ParameterConversion[Value] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


def Field[Value](
    *,
    aliased: str | None = None,
    description: str | Missing = MISSING,
    default: Value | Missing = MISSING,
    default_factory: Callable[[], Value] | Missing = MISSING,
    validator: ParameterValidation[Value] | Missing = MISSING,
    verifier: ParameterVerification[Value] | Missing = MISSING,
    converter: ParameterConversion[Value] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value:  # it is actually a DataField, but type checker has to be fooled
    assert (  # nosec: B101
        default is MISSING or default_factory is MISSING
    ), "Can't specify both default value and factory"
    assert (  # nosec: B101
        description is MISSING or specification is MISSING
    ), "Can't specify both description and specification"
    assert (  # nosec: B101
        validator is MISSING or verifier is MISSING
    ), "Can't specify both validator and verifier"

    return cast(
        Value,
        DataField(
            aliased=aliased,
            description=description,
            default_value=default,
            default_factory=default_factory,
            validator=validator,
            verifier=verifier,
            converter=converter,
            specification=specification,
        ),
    )


@final
class DataField:
    def __init__(
        self,
        aliased: str | None,
        description: str | Missing,
        default_value: Any | Missing,
        default_factory: Callable[[], Any] | Missing,
        validator: ParameterValidation[Any] | Missing,
        verifier: ParameterVerification[Any] | Missing,
        converter: ParameterConversion[Any] | Missing,
        specification: ParameterSpecification | Missing,
    ) -> None:
        self.aliased: str | None = aliased
        self.description: str | Missing = description
        self.default_value: Any | Missing = default_value
        self.default_factory: Callable[[], Any] | Missing = default_factory
        self.validator: ParameterValidation[Any] | Missing = validator
        self.verifier: ParameterVerification[Any] | Missing = verifier
        self.converter: ParameterConversion[Any] | Missing = converter
        self.specification: ParameterSpecification | Missing = specification


def _resolve_field(
    attribute: AttributeAnnotation,
    /,
    *,
    name: str,
    default: Any,
) -> Parameter[Any]:
    match default:
        case DataField() as data_field:
            return Parameter[Any].of(
                attribute,
                name=name,
                alias=data_field.aliased,
                description=data_field.description,
                default=DefaultValue(
                    data_field.default_value,
                    factory=data_field.default_factory,
                ),
                validator=data_field.validator,
                verifier=data_field.verifier,
                converter=data_field.converter,
                specification=data_field.specification,
                required=attribute.required
                and data_field.default_value is MISSING
                and data_field.default_factory is MISSING,
            )

        case DefaultValue() as default:  # pyright: ignore[reportUnknownVariableType]
            return Parameter[Any].of(
                attribute,
                name=name,
                alias=None,
                description=MISSING,
                default=default,  # pyright: ignore[reportUnknownArgumentType]
                validator=MISSING,
                verifier=MISSING,
                converter=MISSING,
                specification=MISSING,
                required=False,
            )

        case value:
            return Parameter[Any].of(
                attribute,
                name=name,
                alias=None,
                description=MISSING,
                default=DefaultValue(value),
                validator=MISSING,
                verifier=MISSING,
                converter=MISSING,
                specification=MISSING,
                required=attribute.required,
            )


@dataclass_transform(
    kw_only_default=True,
    frozen_default=True,
    field_specifiers=(Field,),
)
class DataModelMeta(type):
    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        type_base: type[Any] | None = None,
        type_parameters: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Any:
        data_type = super().__new__(
            cls,
            name,
            bases,
            namespace,
            **kwargs,
        )

        parameters: dict[str, Parameter[Any]] = {}
        parameters_specification: dict[str, ParameterSpecification] = {}
        parameters_specification_required: list[str] = []
        for key, annotation in attribute_annotations(
            data_type,
            type_parameters=type_parameters or {},
        ).items():
            default: Any = getattr(data_type, key, MISSING)
            updated_annotation = annotation.update_required(_check_required(default))
            parameter: Parameter[Any] = _resolve_field(
                updated_annotation,
                name=key,
                default=default,
            )
            # we are using aliased name for specification
            aliased_name: str = parameter.alias or parameter.name
            parameters_specification[aliased_name] = parameter.specification

            if parameter.required:
                parameters_specification_required.append(aliased_name)

            # we are using actual key/name for object itself
            parameters[key] = parameter

        if bases:
            data_type.__PARAMETERS__ = parameters  # pyright: ignore[reportAttributeAccessIssue]
            data_type.__TYPE_PARAMETERS__ = type_parameters  # pyright: ignore[reportAttributeAccessIssue]

        else:
            assert len(parameters) == 0  # nosec: B101
            data_type.__PARAMETERS__ = MISSING  # pyright: ignore[reportAttributeAccessIssue]
            data_type.__TYPE_PARAMETERS__ = None  # pyright: ignore[reportAttributeAccessIssue]

        if parameters_specification:
            data_type.__PARAMETERS_SPECIFICATION__ = {  # pyright: ignore[reportAttributeAccessIssue]
                "type": "object",
                "properties": parameters_specification,
                "required": parameters_specification_required,
            }

        else:
            data_type.__PARAMETERS_SPECIFICATION__ = {  # pyright: ignore[reportAttributeAccessIssue]
                "type": "object",
                "additionalProperties": True,
            }

        data_type.__slots__ = frozenset(parameters.keys())  # pyright: ignore[reportAttributeAccessIssue]
        data_type.__match_args__ = data_type.__slots__  # pyright: ignore[reportAttributeAccessIssue]
        data_type._ = AttributePath(data_type, attribute=data_type)  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue, reportCallIssue]

        return data_type

    def __instancecheck__(
        self,
        instance: Any,
    ) -> bool:
        # check for type match
        if self.__subclasscheck__(type(instance)):  # pyright: ignore[reportUnknownArgumentType]
            return True

        # otherwise check if we are dealing with unparametrized base
        # against the parametrized one, our generic subtypes have base of unparametrized type
        if type(instance) not in self.__bases__:
            return False

        try:
            # validate instance to check unparametrized fields
            _ = self(**vars(instance))

        except Exception:
            return False

        else:
            return True

    def __subclasscheck__(  # noqa: C901, PLR0911, PLR0912
        self,
        subclass: type[Any],
    ) -> bool:
        # check if we are the same class for early exit
        if self == subclass:
            return True

        # then check if we are parametrized
        checked_parameters: Mapping[str, Any] | None = getattr(
            self,
            "__TYPE_PARAMETERS__",
            None,
        )
        if checked_parameters is None:
            # if we are not parametrized allow any subclass
            return self in subclass.__bases__  # pyright: ignore[reportUnnecessaryContains]

        # verify if we have common base next - our generic subtypes have the same base
        if self.__bases__ == subclass.__bases__:
            # if we have the same bases we have different generic subtypes
            # we can verify all of the attributes to check if we have common base
            available_parameters: Mapping[str, Any] | None = getattr(
                subclass,
                "__TYPE_PARAMETERS__",
                None,
            )

            if available_parameters is None:
                # if we have no parameters at this stage this is a serious bug
                raise RuntimeError("Invalid type parametrization for %s", subclass)

            for key, param in checked_parameters.items():
                match available_parameters.get(key):
                    case None:  # if any parameter is missing we should not be there already
                        return False

                    case typing.Any:
                        continue  # Any ignores type checks

                    case checked:
                        if param is Any:
                            continue  # Any ignores type checks

                        elif issubclass(checked, param):
                            continue  # if we have matching type we are fine

                        else:
                            return False  # types are not matching

            return True  # when all parameters were matching we have matching subclass

        elif subclass in self.__bases__:  # our generic subtypes have base of unparametrized type
            # if subclass parameters were not provided then we can be valid ony if all were Any
            return all(param is Any for param in checked_parameters.values())

        else:
            return False  # we have different base / comparing to not parametrized


def _check_required(default: Any) -> bool:
    if default is MISSING:
        return True

    elif isinstance(default, DataField):
        return default.default_value is MISSING and default.default_factory is MISSING

    else:
        return False


_types_cache: WeakValueDictionary[
    tuple[
        Any,
        tuple[Any, ...],
    ],
    Any,
] = WeakValueDictionary()


class DataModel(metaclass=DataModelMeta):
    _: ClassVar[Self]
    __IMMUTABLE__: ClassVar[EllipsisType] = ...
    __PARAMETERS__: ClassVar[dict[str, Parameter[Any]]]
    __PARAMETERS_SPECIFICATION__: ClassVar[ParametersSpecification]
    __TYPE_PARAMETERS__: ClassVar[Mapping[str, Any] | None] = None

    def __class_getitem__(
        cls,
        type_argument: tuple[type[Any], ...] | type[Any],
    ) -> type[Self]:
        assert Generic in cls.__bases__, "Can't specialize non generic type!"  # nosec: B101
        assert cls.__TYPE_PARAMETERS__ is None, "Can't specialize already specialized type!"  # nosec: B101

        type_arguments: tuple[type[Any], ...]
        match type_argument:
            case [*arguments]:
                type_arguments = tuple(arguments)

            case argument:
                type_arguments = (argument,)

        if any(isinstance(argument, TypeVar) for argument in type_arguments):  # pyright: ignore[reportUnnecessaryIsInstance]
            # if we got unfinished type treat it as an alias instead of resolving
            return cast(type[Self], GenericAlias(cls, type_arguments))

        assert len(type_arguments) == len(  # nosec: B101
            cls.__type_params__
        ), "Type arguments count has to match type parameters count"

        if cached := _types_cache.get((cls, type_arguments)):
            return cached

        type_parameters: dict[str, Any] = {
            parameter.__name__: argument
            for (parameter, argument) in zip(
                cls.__type_params__ or (),
                type_arguments or (),
                strict=False,
            )
        }

        parameter_names: str = ",".join(
            getattr(
                argument,
                "__name__",
                str(argument),
            )
            for argument in type_arguments
        )
        name: str = f"{cls.__name__}[{parameter_names}]"
        bases: tuple[type[Self]] = (cls,)

        parametrized_type: type[Self] = DataModelMeta.__new__(
            cls.__class__,
            name=name,
            bases=bases,
            namespace={"__module__": cls.__module__},
            type_parameters=type_parameters,
        )
        _types_cache[(cls, type_arguments)] = parametrized_type
        return parametrized_type

    def __init__(
        self,
        **kwargs: Any,
    ) -> None:
        if self.__PARAMETERS__ is MISSING:
            for key, value in kwargs.items():
                object.__setattr__(
                    self,  # pyright: ignore[reportUnknownArgumentType]
                    key,
                    value,
                )

        else:
            with ParameterValidationContext().scope(self.__class__.__qualname__) as context:
                for name, parameter in self.__PARAMETERS__.items():
                    object.__setattr__(
                        self,  # pyright: ignore[reportUnknownArgumentType]
                        name,
                        parameter.validated(
                            parameter.find(kwargs),
                            context=context,
                        ),
                    )

    @classmethod
    def model_validator(
        cls,
        *,
        verifier: ParameterVerification[Self] | None,
    ) -> ParameterValidation[Self]:
        if verifier := verifier:

            def validator(
                value: Any,
                /,
                context: ParameterValidationContext,
            ) -> Self:
                match value:
                    case validated if isinstance(validated, cls):
                        verifier(validated)
                        return validated

                    case {**values}:
                        validated: Self = cls(**values)
                        verifier(validated)
                        return validated

                    case _:
                        raise TypeError(
                            f"Expected '{cls.__name__}', received '{type(value).__name__}'"
                        )
        else:

            def validator(
                value: Any,
                /,
                context: ParameterValidationContext,
            ) -> Self:
                match value:
                    case valid if isinstance(valid, cls):
                        return valid

                    case {**values}:
                        return cls(**values)

                    case _:
                        raise TypeError(
                            f"Expected '{cls.__name__}', received '{type(value).__name__}'"
                        )

        return validator

    @classmethod
    def from_mapping(
        cls,
        value: Mapping[str, Any],
        /,
    ) -> Self:
        return cls(**value)

    def to_str(self) -> str:
        return self.__str__()

    def to_mapping(
        self,
        aliased: bool = True,
    ) -> Mapping[str, Any]:
        return _data_dict(
            self,
            aliased=aliased,
            converter=None,
        )

    def updating[Value](
        self,
        path: AttributePath[Self, Value] | Value,
        /,
        value: Value,
    ) -> Self:
        assert isinstance(  # nosec: B101
            path, AttributePath
        ), "Prepare parameter path by using Self._.path.to.property or explicitly"

        return cast(AttributePath[Self, Value], path)(self, updated=value)

    def updated(
        self,
        **kwargs: Any,
    ) -> Self:
        return self.__replace__(**kwargs)

    def __replace__(
        self,
        **parameters: Any,
    ) -> Self:
        if not parameters or parameters.keys().isdisjoint(self.__PARAMETERS__.keys()):
            return self  # do not make a copy when nothing will be updated

        with ParameterValidationContext().scope(self.__class__.__qualname__) as context:
            updated: Self = self.__new__(self.__class__)
            for parameter in self.__PARAMETERS__.values():
                validated_value: Any
                if parameter.name in parameters:
                    validated_value = parameter.validated(
                        parameters[parameter.name],
                        context=context,
                    )

                elif parameter.alias and parameter.alias in parameters:
                    validated_value = parameter.validated(
                        parameters[parameter.alias],
                        context=context,
                    )

                else:  # no need to validate again when reusing current value
                    validated_value = object.__getattribute__(
                        self,
                        parameter.name,
                    )

                object.__setattr__(
                    updated,
                    parameter.name,
                    validated_value,
                )

            return updated

    def __str__(self) -> str:
        return _data_str(self, aliased=True, converter=None).strip()

    def __repr__(self) -> str:
        return str(self.to_mapping())

    def __eq__(self, other: Any) -> bool:
        if other.__class__ != self.__class__:
            return False

        if self.__PARAMETERS__ is MISSING:
            return all(value == getattr(other, key, MISSING) for key, value in vars(self).items())

        else:
            return all(
                getattr(self, key, MISSING) == getattr(other, key, MISSING)
                for key in self.__PARAMETERS__.keys()
            )

    def __contains__(
        self,
        element: Any,
    ) -> bool:
        return element in vars(self)

    def __getitem__(
        self,
        name: str,
    ) -> Any | Missing:
        return vars(self).get(name, MISSING)

    @overload
    def get(
        self,
        name: str,
    ) -> Any | Missing: ...

    @overload
    def get[Default](
        self,
        name: str,
        default: Default,
    ) -> Any | Default: ...

    def get(
        self,
        name: str,
        default: Any | Missing = MISSING,
    ) -> Any | Missing:
        if self.__contains__(name):
            return self.__getitem__(name)

        else:
            return default

    def __setattr__(
        self,
        __name: str,
        __value: Any,
    ) -> None:
        raise AttributeError(f"{self.__class__.__qualname__} is frozen and can't be modified")

    def __delattr__(
        self,
        __name: str,
    ) -> None:
        raise AttributeError(f"{self.__class__.__qualname__} is frozen and can't be modified")

    def __iter__(self) -> Iterator[str]:
        yield from vars(self)

    def __len__(self) -> int:
        return len(vars(self))

    def __copy__(self) -> Self:
        return self  # DataModel is immutable, no need to provide an actual copy

    def __deepcopy__(
        self,
        memo: dict[int, Any] | None,
    ) -> Self:
        return self  # DataModel is immutable, no need to provide an actual copy

    @classmethod
    def json_schema(
        cls,
        indent: int | None = None,
    ) -> str:
        return json_schema(
            cls.__PARAMETERS_SPECIFICATION__,
            indent=indent,
        )

    @classmethod
    def simplified_schema(
        cls,
        indent: int | None = None,
    ) -> str:
        assert not_missing(  # nosec: B101
            cls.__PARAMETERS_SPECIFICATION__
        ), f"{cls.__qualname__} can't be represented using simplified schema"

        return simplified_schema(
            cls.__PARAMETERS_SPECIFICATION__,
            indent=indent,
        )

    @classmethod
    def from_json(
        cls,
        value: str | bytes,
        /,
        decoder: type[json.JSONDecoder] = json.JSONDecoder,
    ) -> Self:
        try:
            return cls(
                **json.loads(
                    value,
                    cls=decoder,
                )
            )

        except Exception as exc:
            raise ValueError(f"Failed to decode {cls.__name__} from json:\n{value}") from exc

    def to_json(
        self,
        aliased: bool = True,
        indent: int | None = None,
        encoder_class: type[json.JSONEncoder] | None = None,
    ) -> str:
        try:
            return json.dumps(
                self.to_mapping(aliased=aliased),
                indent=indent,
                cls=encoder_class or ModelJSONEncoder,
            )

        except Exception as exc:
            raise ValueError(
                f"Failed to encode {self.__class__.__name__} to json:\n{self.to_mapping()}"
            ) from exc


class ModelJSONEncoder(json.JSONEncoder):
    def default(self, o: object) -> Any:
        if isinstance(o, UUID):
            return o.hex

        elif isinstance(o, datetime):
            return o.isoformat()

        elif isinstance(o, time):
            return o.isoformat()

        elif isinstance(o, date):
            return o.isoformat()

        else:
            return json.JSONEncoder.default(self, o)


# based on python dataclass asdict but simplified
def _data_dict(  # noqa: PLR0911
    data: Any,
    /,
    aliased: bool,
    converter: ParameterConversion[Any] | None,
) -> Any:
    # use converter if able
    if converter := converter:
        return converter(data)

    match data:
        case str() | None | int() | float() | bool():
            return data  # use basic value types as they are

        case parametrized_data if hasattr(parametrized_data, "__PARAMETERS__"):
            # convert parametrized data to dict
            if parametrized_data.__PARAMETERS__ is MISSING:  # if base - use all variables
                return cast(
                    dict[str, Any],
                    {
                        key: value
                        for key, value in vars(parametrized_data).items()  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]
                        if value is not MISSING
                    },
                )

            elif aliased:  # alias if needed
                return {
                    field.alias or field.name: _data_dict(
                        getattr(parametrized_data, field.name),
                        aliased=aliased,
                        converter=field.converter,
                    )
                    for field in parametrized_data.__PARAMETERS__.values()
                    if getattr(parametrized_data, field.name, MISSING) is not MISSING
                }

            else:
                return {
                    field.name: _data_dict(
                        getattr(parametrized_data, field.name),
                        aliased=aliased,
                        converter=field.converter,
                    )
                    for field in parametrized_data.__PARAMETERS__.values()
                    if getattr(parametrized_data, field.name, MISSING) is not MISSING
                }

        case {**elements}:  # replace mapping with dict
            return {
                key: _data_dict(
                    value,
                    aliased=aliased,
                    converter=None,
                )
                for key, value in elements.items()
                if value is not MISSING
            }

        case [*values]:  # replace sequence with list
            return [
                _data_dict(
                    value,
                    aliased=aliased,
                    converter=None,
                )
                for value in values
                if value is not MISSING
            ]

        case dataclass if is_dataclass(dataclass):
            return {
                f.name: _data_dict(
                    getattr(dataclass, f.name),
                    aliased=aliased,
                    converter=None,
                )
                for f in dataclass_fields(dataclass)
            }

        case other:  # for other types use deepcopy
            return deepcopy(other)


def _data_str(  # noqa: PLR0911, PLR0912, C901
    data: Any,
    /,
    aliased: bool,
    converter: ParameterConversion[Any] | None,
) -> str:
    # use converter if able
    if converter := converter:
        return _data_str(
            converter(data),
            aliased=aliased,
            converter=None,
        )

    match data:
        case str() as string:
            return string

        case None | int() | float() | bool():
            return str(data)

        case parametrized_data if hasattr(parametrized_data, "__PARAMETERS__"):
            # convert parametrized data to dict
            if parametrized_data.__PARAMETERS__ is MISSING:  # if base - use all variables
                string: str = ""
                for key, value in vars(parametrized_data).items():  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
                    element: str = _data_str(
                        value,
                        aliased=aliased,
                        converter=None,
                    ).replace("\n", "\n  ")

                    string += f"\n{key}: {element}"

                return string

            elif aliased:  # alias if needed
                string: str = ""
                for field in parametrized_data.__PARAMETERS__.values():
                    element: str = _data_str(
                        getattr(parametrized_data, field.name),
                        aliased=aliased,
                        converter=field.converter,
                    ).replace("\n", "\n  ")

                    string += f"\n{field.alias or field.name}: {element}"

                return string

            else:
                string: str = ""
                for field in parametrized_data.__PARAMETERS__.values():
                    element: str = _data_str(
                        getattr(parametrized_data, field.name),
                        aliased=aliased,
                        converter=field.converter,
                    ).replace("\n", "\n  ")

                    string += f"\n{field.name}: {element}"

                return string

        case {**elements}:  # replace mapping with dict
            string: str = ""
            for key, value in elements.items():
                element: str = _data_str(
                    value,
                    aliased=aliased,
                    converter=None,
                ).replace("\n", "\n  ")

                string += f"\n{key}: {element}"

            return string

        case [*values]:  # replace sequence with list
            string: str = ""
            for value in values:
                element: str = (
                    _data_str(
                        value,
                        aliased=aliased,
                        converter=None,
                    )
                    .replace("\n", "\n  ")
                    .strip()
                )

                string += f"\n- {element}"

            return string

        case dataclass if is_dataclass(dataclass):
            string: str = ""
            for field in dataclass_fields(dataclass):
                element: str = _data_str(
                    getattr(parametrized_data, field.name),
                    aliased=aliased,
                    converter=None,
                ).replace("\n", "\n  ")

                string += f"\n{field.name}: {element}"

            return string

        case other:  # for other types use its str
            return str(other)



from asyncio import gather
from collections.abc import Callable, Sequence
from typing import Any, Protocol, Self, cast, overload, runtime_checkable

from haiway import AttributePath, ScopeContext, ctx

from draive.commons import META_EMPTY, Meta, MetaValues
from draive.evaluation.evaluator import EvaluatorResult, PreparedEvaluator
from draive.parameters import DataModel, Field

__all__ = (
    "EvaluationScenarioResult",
    "ScenarioEvaluator",
    "ScenarioEvaluatorDefinition",
    "ScenarioEvaluatorResult",
    "evaluation_scenario",
)


class ScenarioEvaluatorResult(DataModel):
    scenario: str = Field(
        description="Name of evaluated scenario",
    )
    evaluations: Sequence[EvaluatorResult] = Field(
        description="Scenario evaluation results",
    )
    meta: Meta = Field(
        description="Additional evaluation metadata",
        default=META_EMPTY,
    )

    @property
    def passed(self) -> bool:
        # empty evaluations is equivalent of failure
        return len(self.evaluations) > 0 and all(case.passed for case in self.evaluations)

    def report(
        self,
        *,
        include_passed: bool = True,
        include_details: bool = True,
    ) -> str:
        evaluations_report: str = "\n".join(
            result.report(include_details=include_details)
            for result in self.evaluations
            if include_passed or not result.passed
        )

        if evaluations_report:  # nonempty report
            if include_details:
                return (
                    f"<scenario name='{self.scenario}'>"
                    f"\n<relative_score>{self.relative_score * 100:.2f}%</relative_score>"
                    f"\n<evaluations>\n{evaluations_report}\n</evaluations>"
                    "\n</scenario>"
                )

            else:
                return f"Scenario {self.scenario}:\n{evaluations_report}"

        elif not self.evaluations:
            return f"Scenario {self.scenario} empty!"

        else:
            return f"Scenario {self.scenario} passed!"

    @property
    def relative_score(self) -> float:
        if not self.evaluations:
            return 0.0

        return len([evaluation for evaluation in self.evaluations if evaluation.passed]) / len(
            self.evaluations
        )


class EvaluationScenarioResult(DataModel):
    @classmethod
    async def evaluating[Value](
        cls,
        value: Value,
        /,
        evaluators: PreparedEvaluator[Value],
        *_evaluators: PreparedEvaluator[Value],
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            evaluations=tuple(
                await gather(
                    *(evaluator(value) for evaluator in (evaluators, *_evaluators)),
                    return_exceptions=False,
                ),
            ),
            meta=Meta.of(meta),
        )

    evaluations: Sequence[EvaluatorResult] = Field(
        description="Scenario evaluation results",
    )
    meta: Meta = Field(
        description="Additional evaluation metadata",
        default=META_EMPTY,
    )


@runtime_checkable
class PreparedScenarioEvaluator[Value](Protocol):
    async def __call__(
        self,
        value: Value,
        /,
    ) -> ScenarioEvaluatorResult: ...


@runtime_checkable
class ScenarioEvaluatorDefinition[Value, **Args](Protocol):
    @property
    def __name__(self) -> str: ...

    async def __call__(
        self,
        value: Value,
        /,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Sequence[EvaluatorResult] | EvaluationScenarioResult: ...


class ScenarioEvaluator[Value, **Args]:
    __slots__ = ("_definition", "_execution_context", "meta", "name")

    def __init__(
        self,
        name: str,
        definition: ScenarioEvaluatorDefinition[Value, Args],
        execution_context: ScopeContext | None,
        meta: Meta,
    ) -> None:
        assert "\n" not in name  # nosec: B101

        self.name: str
        object.__setattr__(
            self,
            "name",
            name.lower().replace(" ", "_"),
        )
        self._definition: ScenarioEvaluatorDefinition[Value, Args]
        object.__setattr__(
            self,
            "_definition",
            definition,
        )
        self._execution_context: ScopeContext | None
        object.__setattr__(
            self,
            "_execution_context",
            execution_context,
        )
        self.meta: Meta
        object.__setattr__(
            self,
            "meta",
            meta,
        )

    def prepared(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> PreparedScenarioEvaluator[Value]:
        async def evaluate(
            value: Value,
        ) -> ScenarioEvaluatorResult:
            return await self(
                value,
                *args,
                **kwargs,
            )

        return evaluate

    def with_name(
        self,
        name: str,
        /,
    ) -> Self:
        return self.__class__(
            name=name,
            definition=self._definition,
            execution_context=self._execution_context,
            meta=self.meta,
        )

    def with_execution_context(
        self,
        context: ScopeContext,
        /,
    ) -> Self:
        return self.__class__(
            name=self.name,
            definition=self._definition,
            execution_context=context,
            meta=self.meta,
        )

    def with_meta(
        self,
        meta: Meta | MetaValues,
        /,
    ) -> Self:
        return self.__class__(
            name=self.name,
            definition=self._definition,
            execution_context=self._execution_context,
            meta=self.meta.merged_with(meta),
        )

    def contra_map[Mapped](
        self,
        mapping: Callable[[Mapped], Value] | AttributePath[Mapped, Value] | Value,
        /,
    ) -> "ScenarioEvaluator[Mapped, Args]":
        mapper: Callable[[Mapped], Value]
        match mapping:
            case Callable() as function:  # pyright: ignore[reportUnknownVariableType]
                mapper = function

            case path:
                assert isinstance(  # nosec: B101
                    path, AttributePath
                ), "Prepare parameter path by using Self._.path.to.property"
                mapper = cast(AttributePath[Mapped, Value], path).__call__

        async def evaluation(
            value: Mapped,
            /,
            *args: Args.args,
            **kwargs: Args.kwargs,
        ) -> Sequence[EvaluatorResult] | EvaluationScenarioResult:
            return await self._definition(
                mapper(value),
                *args,
                **kwargs,
            )

        return ScenarioEvaluator[Mapped, Args](
            name=self.name,
            definition=evaluation,
            execution_context=self._execution_context,
            meta=self.meta,
        )

    async def __call__(
        self,
        value: Value,
        /,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> ScenarioEvaluatorResult:
        result: ScenarioEvaluatorResult
        if context := self._execution_context:
            async with context:
                result = await self._evaluate(
                    value,
                    *args,
                    **kwargs,
                )

        else:
            async with ctx.scope(f"evaluation.scenario.{self.name}"):
                result = await self._evaluate(
                    value,
                    *args,
                    **kwargs,
                )

        ctx.record(
            metric=f"evaluation.scenario.{result.scenario}.score",
            value=result.relative_score,
            unit="%",
            attributes={
                "evaluation.scenario.evaluations": [
                    evaluation.evaluator for evaluation in result.evaluations
                ],
                "evaluation.passed": result.passed,
            },
        )
        return result

    async def _evaluate(
        self,
        value: Value,
        /,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> ScenarioEvaluatorResult:
        try:
            match await self._definition(
                value,
                *args,
                **kwargs,
            ):
                case EvaluationScenarioResult() as result:
                    meta: Meta
                    if self.meta:
                        if result.meta:
                            meta = self.meta.merged_with(result.meta)

                        else:
                            meta = self.meta

                    else:
                        meta = result.meta

                    return ScenarioEvaluatorResult(
                        scenario=self.name,
                        evaluations=result.evaluations,
                        meta=meta,
                    )

                case [*results]:
                    return ScenarioEvaluatorResult(
                        scenario=self.name,
                        evaluations=tuple(results),
                        meta=self.meta,
                    )

        except Exception as exc:
            ctx.log_error(
                f"Scenario evaluator `{self.name}` failed, using empty fallback result",
                exception=exc,
            )

            return ScenarioEvaluatorResult(
                scenario=self.name,
                evaluations=(),
                meta=self.meta.updated(exception=str(exc)),
            )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )


@overload
def evaluation_scenario[Value, **Args](
    definition: ScenarioEvaluatorDefinition[Value, Args],
    /,
) -> ScenarioEvaluator[Value, Args]: ...


@overload
def evaluation_scenario[Value, **Args](
    *,
    name: str | None = None,
    execution_context: ScopeContext | None = None,
    meta: Meta | MetaValues | None = None,
) -> Callable[
    [ScenarioEvaluatorDefinition[Value, Args]],
    ScenarioEvaluator[Value, Args],
]: ...


def evaluation_scenario[Value, **Args](  # pyright: ignore[reportInconsistentOverload] - this seems to be pyright false positive/error
    definition: ScenarioEvaluatorDefinition[Value, Args] | None = None,
    /,
    *,
    name: str | None = None,
    execution_context: ScopeContext | None = None,
    meta: Meta | MetaValues | None = None,
) -> (
    Callable[
        [ScenarioEvaluatorDefinition[Value, Args]],
        ScenarioEvaluator[Value, Args],
    ]
    | ScenarioEvaluator[Value, Args]
):
    def wrap(
        definition: ScenarioEvaluatorDefinition[Value, Args],
    ) -> ScenarioEvaluator[Value, Args]:
        return ScenarioEvaluator(
            name=name or definition.__name__,
            definition=definition,
            execution_context=execution_context,
            meta=Meta.of(meta),
        )

    if definition:
        return wrap(definition)

    else:
        return wrap



from typing import Any, Self

from draive.evaluation.value import EvaluationScoreValue, evaluation_score_value
from draive.parameters import DataModel, Field

__all__ = ("EvaluationScore",)


def _verifier(
    value: float,
) -> None:
    if not (0 <= value <= 1):
        raise ValueError(f"Score has to have a value between 0 and 1, received: {value}")


class EvaluationScore(DataModel):
    @classmethod
    def of(
        cls,
        score: EvaluationScoreValue,
        /,
        *,
        comment: str | None = None,
    ) -> Self:
        return cls(
            value=evaluation_score_value(score),
            comment=comment,
        )

    value: float = Field(
        description="Score value, between 0 (failure) and 1 (success)",
        verifier=_verifier,
    )
    comment: str | None = Field(
        description="Explanation of the score",
        default=None,
    )

    def __eq__(self, other: Any) -> bool:
        match other:
            case float() as value:
                return self.value == value

            case EvaluationScore() as score:
                return self.value == score.value

            case _:
                return NotImplemented

    def __ne__(self, other: Any) -> bool:
        match other:
            case float() as value:
                return self.value != value

            case EvaluationScore() as score:
                return self.value != score.value

            case _:
                return NotImplemented

    def __lt__(self, other: Any) -> bool:
        match other:
            case float() as value:
                return self.value < value

            case EvaluationScore() as score:
                return self.value < score.value

            case _:
                return NotImplemented

    def __le__(self, other: Any) -> bool:
        match other:
            case float() as value:
                return self.value <= value

            case EvaluationScore() as score:
                return self.value <= score.value

            case _:
                return NotImplemented

    def __gt__(self, other: Any) -> bool:
        match other:
            case float() as value:
                return self.value > value

            case EvaluationScore() as score:
                return self.value > score.value

            case _:
                return NotImplemented

    def __ge__(self, other: Any) -> bool:
        match other:
            case float() as value:
                return self.value >= value

            case EvaluationScore() as score:
                return self.value >= score.value

            case _:
                return NotImplemented



from typing import Any

from haiway import State, ctx

from draive.generation.image.typing import ImageGenerating
from draive.instructions import Instruction
from draive.multimodal import MediaContent, Multimodal

__all__ = ("ImageGeneration",)


class ImageGeneration(State):
    @classmethod
    async def generate(
        cls,
        *,
        instruction: Instruction | str,
        input: Multimodal | None = None,  # noqa: A002
        **extra: Any,
    ) -> MediaContent:
        return await ctx.state(ImageGeneration).generate(
            instruction=instruction,
            input=input,
            **extra,
        )

    generating: ImageGenerating



from collections.abc import Collection, Iterable
from types import TracebackType
from typing import Literal, final, overload

from haiway import State

from draive.cohere.api import CohereAPI
from draive.cohere.embedding import CohereEmbedding

__all__ = ("Cohere",)


@final
class Cohere(
    CohereEmbedding,
    CohereAPI,
):
    """
    Access to Cohere services, can be used to prepare various functionalities like embedding.
    """

    __slots__ = ("_features",)

    @overload
    def __init__(
        self,
        provider: Literal["bedrock"],
        /,
        *,
        aws_region: str | None = None,
        timeout: float = 60.0,
        features: Collection[Literal["text_embedding", "image_embedding"]] | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        provider: Literal["cohere"] = "cohere",
        /,
        *,
        base_url: str | None = None,
        api_key: str | None = None,
        timeout: float = 60.0,
        features: Collection[Literal["text_embedding", "image_embedding"]] | None = None,
    ) -> None: ...

    def __init__(
        self,
        provider: Literal["cohere", "bedrock"] = "cohere",
        /,
        *,
        base_url: str | None = None,
        api_key: str | None = None,
        aws_region: str | None = None,
        timeout: float = 60.0,
        features: Collection[Literal["text_embedding", "image_embedding"]] | None = None,
    ) -> None:
        super().__init__(
            provider,
            base_url=base_url,
            api_key=api_key,
            aws_region=aws_region,
            timeout=timeout,
        )

        self._features: frozenset[Literal["text_embedding", "image_embedding"]] = (
            frozenset(features) if features is not None else frozenset(("text_embedding",))
        )

    async def __aenter__(self) -> Iterable[State]:
        state: list[State] = []

        if "text_embedding" in self._features:
            state.append(self.text_embedding())

        if "image_embedding" in self._features:
            state.append(self.image_embedding())

        return state

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._deinitialize_client()



from collections.abc import Iterable
from typing import Any, Literal, Protocol, Self, overload, runtime_checkable
from uuid import UUID, uuid4

from haiway import Default, MissingState, State, ctx

from draive.commons import META_EMPTY, Meta, MetaValues
from draive.multimodal import MultimodalContent
from draive.multimodal.content import Multimodal
from draive.parameters import DataModel

__all__ = (
    "Processing",
    "ProcessingEvent",
    "ProcessingState",
)


class ProcessingEvent(DataModel):
    @classmethod
    def of(
        cls,
        name: str,
        /,
        *,
        identifier: UUID | None = None,
        content: Multimodal | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            identifier=identifier if identifier is not None else uuid4(),
            name=name,
            content=MultimodalContent.of(content) if content is not None else None,
            meta=Meta.of(meta),
        )

    identifier: UUID = Default(factory=uuid4)
    name: str
    content: MultimodalContent | None = None
    meta: Meta = META_EMPTY


@runtime_checkable
class ProcessingEventReporting(Protocol):
    async def __call__(
        self,
        event: ProcessingEvent,
        /,
    ) -> None: ...


@runtime_checkable
class ProcessingStateReading(Protocol):
    async def __call__[StateType: DataModel | State](
        self,
        state: type[StateType],
        /,
    ) -> StateType | None: ...


@runtime_checkable
class ProcessingStateWriting(Protocol):
    async def __call__(
        self,
        value: DataModel | State,
        /,
    ) -> None: ...


async def _log_event_reporting(
    event: ProcessingEvent,
    /,
) -> None:
    ctx.log_info(f"ProcessingEvent:\n{event}")


async def _context_state_reading[StateType: DataModel | State](
    state: type[StateType],
    /,
    default: StateType | None = None,
) -> StateType | None:
    if issubclass(state, State):
        return ctx.state(state, default=default)

    else:
        return None


async def _ignored_state_writing(
    state: DataModel | State,
    /,
) -> None:
    ctx.log_warning(f"Ignoring processing state writing of {type(state)}")


class Processing(State):
    @overload
    @classmethod
    async def read[StateType: DataModel | State](
        cls,
        state: type[StateType],
        /,
    ) -> StateType | None: ...

    @overload
    @classmethod
    async def read[StateType: DataModel | State](
        cls,
        state: type[StateType],
        /,
        *,
        required: Literal[True],
    ) -> StateType: ...

    @overload
    @classmethod
    async def read[StateType: DataModel | State](
        cls,
        state: type[StateType],
        /,
        *,
        default: StateType,
    ) -> StateType: ...

    @classmethod
    async def read[StateType: DataModel | State](
        cls,
        state: type[StateType],
        /,
        *,
        default: StateType | None = None,
        required: bool = False,
    ) -> StateType | None:
        current: StateType | None = await ctx.state(cls).state_reading(state)
        if current is not None:
            return current

        elif default is not None:
            return default

        elif required:
            raise MissingState(f"{state.__qualname__} is not available within current processing")

        else:
            return None

    @classmethod
    async def write(
        cls,
        state: DataModel | State,
        /,
    ) -> None:
        await ctx.state(cls).state_writing(state)

    @overload
    @classmethod
    async def event(
        cls,
        event: ProcessingEvent,
        /,
    ) -> None: ...

    @overload
    @classmethod
    async def event(
        cls,
        /,
        *,
        identifier: UUID | None = None,
        name: str | None,
        content: Multimodal | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> None: ...

    @classmethod
    async def event(
        cls,
        event: ProcessingEvent | None = None,
        /,
        *,
        identifier: UUID | None = None,
        name: str | None = None,
        content: Multimodal | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> None:
        if event is not None:
            assert identifier is None and name is None  # nosec: B101
            return await ctx.state(cls).event_reporting(event)

        assert name is not None  # nosec: B101
        return await ctx.state(cls).event_reporting(
            ProcessingEvent.of(
                name,
                identifier=identifier,
                content=MultimodalContent.of(content) if content is not None else None,
                meta=Meta.of(meta),
            )
        )

    event_reporting: ProcessingEventReporting = _log_event_reporting
    state_reading: ProcessingStateReading = _context_state_reading
    state_writing: ProcessingStateWriting = _ignored_state_writing


class ProcessingState:
    def __init__(
        self,
        state: Iterable[DataModel | State] | None,
    ) -> None:
        self._state: dict[type[DataModel | State], Any]
        object.__setattr__(
            self,
            "_state",
            {type(element): element for element in state} if state else {},
        )

    async def read[StateType: DataModel | State](
        self,
        state: type[StateType],
        /,
        default: StateType | None = None,
    ) -> StateType | None:
        if state in self._state:
            return self._state[state]

        else:
            return default

    async def write(
        self,
        state: DataModel | State,
        /,
    ) -> None:
        self._state[type(state)] = state

    def __setattr__(
        self,
        __name: str,
        __value: Any,
    ) -> None:
        raise RuntimeError("ProcessingState is frozen and can't be modified")

    def __delattr__(
        self,
        __name: str,
    ) -> None:
        raise RuntimeError("ProcessingState is frozen and can't be modified")



from draive.generation.model.state import ModelGeneration
from draive.generation.model.types import ModelGenerating, ModelGeneratorDecoder

__all__ = (
    "ModelGenerating",
    "ModelGeneration",
    "ModelGeneratorDecoder",
)



from collections.abc import Sequence
from typing import Any

from draive.utils.memory import Memory

__all__ = (
    "AccumulativeVolatileMemory",
    "ConstantMemory",
    "VolatileMemory",
)


def ConstantMemory[Recalled](
    recalled: Recalled,
) -> Memory[Recalled, Any]:
    async def recall(
        **extra: Any,
    ) -> Recalled:
        return recalled

    async def remember(
        *items: Any,
        **extra: Any,
    ) -> None:
        pass  # noop

    return Memory[Recalled, Any](
        recall=recall,
        remember=remember,
    )


def VolatileMemory[Item](
    *,
    initial: Item,
    limit: int | None = None,
) -> Memory[Item, Item]:
    storage: Item = initial
    limit = limit or 0

    async def recall(
        **extra: Any,
    ) -> Item:
        nonlocal storage
        return storage

    async def remember(
        *items: Item,
        **extra: Any,
    ) -> None:
        nonlocal storage
        if not items:
            return  # nothing to do

        storage = items[-1]

    return Memory[Item, Item](
        recall=recall,
        remember=remember,
    )


def AccumulativeVolatileMemory[Item](
    *,
    initial: Sequence[Item] | None = None,
    limit: int | None = None,
) -> Memory[Sequence[Item], Item]:
    storage: Sequence[Item] = tuple(initial) if initial else ()
    limit = limit or 0

    async def recall(
        **extra: Any,
    ) -> Sequence[Item]:
        nonlocal storage
        return storage

    async def remember(
        *items: Item,
        **extra: Any,
    ) -> None:
        nonlocal storage
        if not items:
            return  # nothing to do

        storage = (*storage, *items)

        if limit > 0:
            storage = tuple(storage[-limit:])

    return Memory[Sequence[Item], Item](
        recall=recall,
        remember=remember,
    )



from asyncio import Lock
from collections.abc import Callable, Iterable, MutableMapping, MutableSequence, Sequence
from typing import Any, cast

from haiway import AttributePath, AttributeRequirement

from draive.embedding import Embedded, ImageEmbedding, TextEmbedding
from draive.multimodal import MediaContent, MediaData, MediaReference, TextContent
from draive.parameters import DataModel
from draive.similarity import mmr_vector_similarity_search, vector_similarity_search
from draive.utils import VectorIndex

__all__ = ("VolatileVectorIndex",)


def VolatileVectorIndex() -> VectorIndex:  # noqa: C901, PLR0915
    lock: Lock = Lock()
    storage: MutableMapping[type[Any], MutableSequence[Embedded[Any]]] = {}

    async def index[Model: DataModel, Value: MediaContent | TextContent | str](  # noqa: C901, PLR0912
        model: type[Model],
        /,
        *,
        attribute: Callable[[Model], Value] | AttributePath[Model, Value] | Value,
        values: Iterable[Model],
        **extra: Any,
    ) -> None:
        async with lock:
            value_selector: Callable[[Model], Value]
            match attribute:
                case Callable() as selector:
                    value_selector = selector

                case path:
                    assert isinstance(  # nosec: B101
                        path, AttributePath
                    ), "Prepare parameter path by using Self._.path.to.property"
                    value_selector = cast(AttributePath[Model, Value], path).__call__

            selected_values: list[str | bytes] = []
            for value in values:
                match value_selector(value):
                    case str() as text:
                        selected_values.append(text)

                    case TextContent() as text_content:
                        selected_values.append(text_content.text)

                    case MediaData() as media_data:
                        if media_data.kind != "image":
                            raise ValueError(f"{media_data.kind} embedding is not supported")

                        selected_values.append(media_data.data)

                    case MediaReference():
                        raise ValueError("Media references are not supported")

            embedded_values: Sequence[Embedded[str] | Embedded[bytes]]
            if all(isinstance(value, str) for value in selected_values):
                embedded_values = await TextEmbedding.embed(
                    cast(list[str], selected_values),
                    **extra,
                )

            elif all(value for value in selected_values):
                embedded_values = await ImageEmbedding.embed(
                    cast(list[bytes], selected_values),
                    **extra,
                )

            else:
                raise ValueError("Selected attribute values have to be the same type")

            embedded_models: list[Embedded[Model]] = [
                Embedded(
                    value=value,
                    vector=embedded.vector,
                )
                for value, embedded in zip(
                    values,
                    embedded_values,
                    strict=True,
                )
            ]

            if model in storage:
                storage[model].extend(embedded_models)

            else:
                storage[model] = embedded_models

    async def search[Model: DataModel](  # noqa: C901, PLR0912
        model: type[Model],
        /,
        *,
        query: Sequence[float] | MediaContent | TextContent | str | None = None,
        score_threshold: float | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int | None = None,
        **extra: Any,
    ) -> Iterable[Model]:
        assert query is not None or (query is None and score_threshold is None)  # nosec: B101
        async with lock:
            stored: Sequence[Embedded[Model]] | None = storage.get(model)
            if not stored:
                return []

            filtered: Sequence[Embedded[Model]]
            if requirements:
                filtered = [
                    element
                    for element in stored
                    if requirements.check(
                        element.value,
                        raise_exception=False,
                    )
                ]

            else:
                filtered = stored

            if not filtered:
                return []

            query_vector: Sequence[float]
            match query:
                case None:
                    if limit:
                        return [embedded.value for embedded in filtered[:limit]]

                    else:
                        return [embedded.value for embedded in filtered]

                case str() as text:
                    embedded_text: Embedded[str] = await TextEmbedding.embed(
                        text,
                        **extra,
                    )
                    query_vector = embedded_text.vector

                case TextContent() as text_content:
                    embedded_text: Embedded[str] = await TextEmbedding.embed(
                        text_content.text,
                        **extra,
                    )
                    query_vector = embedded_text.vector

                case MediaData() as media_data:
                    if media_data.kind != "image":
                        raise ValueError(f"{media_data.kind} embedding is not supported")

                    embedded_image: Embedded[bytes] = await ImageEmbedding.embed(
                        media_data.data,
                        **extra,
                    )
                    query_vector = embedded_image.vector

                case MediaReference():
                    raise ValueError("Media references are not supported")

                case vector:
                    query_vector = vector

            matching: Sequence[Embedded[Model]] = [
                filtered[index]
                for index in vector_similarity_search(
                    query_vector=query_vector,
                    values_vectors=[element.vector for element in filtered],
                    score_threshold=score_threshold,
                    limit=limit * 8  # feed MMR with more results
                    if limit is not None
                    else None,
                )
            ]

            if not matching:
                return []

            return [
                matching[index].value
                for index in mmr_vector_similarity_search(
                    query_vector=query_vector,
                    values_vectors=[element.vector for element in matching],
                    limit=limit,
                )
            ]

    async def delete[Model: DataModel](
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        **extra: Any,
    ) -> None:
        async with lock:
            if model not in storage:
                return

            if requirements is not None:
                storage[model] = [
                    element
                    for element in storage[model]
                    if not requirements.check(
                        element.value,
                        raise_exception=False,
                    )
                ]

            else:
                del storage[model]

    return VectorIndex(
        indexing=index,
        searching=search,
        deleting=delete,
    )



#! uv -n run
# /// script
# requires-python = ">=3.12"
# dependencies = [
#   "draive[gemini]~=0.57",
#   "pyaudio~=0.2",
#   "beautifulsoup4~=4.13",
#   "httpx~=0.28",
# ]
# ///
from asyncio import run, to_thread
from collections.abc import AsyncGenerator

import pyaudio
from bs4 import BeautifulSoup
from haiway import MetricsLogger
from haiway.utils.logs import setup_logging
from haiway.utils.queue import AsyncQueue
from httpx import AsyncClient, Response

from draive import (
    Instruction,
    MediaData,
    Realtime,
    ctx,
    load_env,
    tool,
)
from draive.gemini import Gemini, GeminiLiveConfig

load_env()
setup_logging("realtime_chat")

pya = pyaudio.PyAudio()
FORMAT = pyaudio.paInt16
CHANNELS = 1
SEND_SAMPLE_RATE = 16000
RECEIVE_SAMPLE_RATE = 24000
CHUNK_SIZE = 1024


@tool(description="Access content of a website")
async def website_content(url: str) -> str:
    # simplified implementation of web access
    async with AsyncClient(follow_redirects=True) as client:
        response: Response = await client.get(url)
        page = BeautifulSoup(
            markup=await response.aread(),
            features="html.parser",
        )
        return (page.find(name="main") or page.find(name="body") or page).text


async def main():
    async with ctx.scope(
        "realtime_chat",
        GeminiLiveConfig(model="gemini-2.0-flash-exp"),
        disposables=(Gemini(http_options={"api_version": "v1alpha"}),),
        metrics=MetricsLogger.handler(),
    ):
        audio_replay_queue = AsyncQueue()

        async def replay_audio() -> None:
            out_audio_stream = await to_thread(
                pya.open,
                format=FORMAT,
                channels=CHANNELS,
                rate=RECEIVE_SAMPLE_RATE,
                output=True,
            )
            async for audio in audio_replay_queue:
                await to_thread(
                    out_audio_stream.write,
                    audio,
                )

        ctx.spawn(replay_audio)

        async def record_audio() -> AsyncGenerator[MediaData]:
            mic_info = pya.get_default_input_device_info()
            audio_stream = await to_thread(
                pya.open,
                format=FORMAT,
                channels=CHANNELS,
                rate=SEND_SAMPLE_RATE,
                input=True,
                input_device_index=mic_info["index"],
                frames_per_buffer=CHUNK_SIZE,
            )

            while True:
                yield MediaData.of(
                    await to_thread(audio_stream.read, CHUNK_SIZE),
                    media="audio/pcm",
                )

        async for chunk in await Realtime.process(
            instruction=Instruction.of(
                "You are helpful assistant."
                " Try to autonomously complete user requests using available tools when needed."
                " Request user help or clarification only on with ambiguity or issues."
            ),
            input=record_audio(),
            tools=[website_content],
        ):
            if chunk.meta(category="interrupted"):
                # on interrupt event clear output queue to prevent playing unnecessary content
                audio_replay_queue.clear()

            for part in chunk.parts:
                match part:
                    case MediaData() as media:
                        audio_replay_queue.enqueue(media.data)

                    case _:
                        pass  # we are expecting only audio outputs


run(main())



from base64 import b64decode, b64encode
from typing import Any

from haiway import MISSING, Missing, not_missing

from draive.parameters import (
    Field,
    ParameterValidationContext,
)

__all__ = ("b64_data_field",)


def b64_data_field(
    *,
    description: str | Missing = MISSING,
) -> Any:
    return Field(
        specification={
            "type": "string",
            "description": description,
        }
        if not_missing(description)
        else {
            "type": "string",
        },
        validator=_b64_validator,
        converter=_b64_converter,
    )


def _b64_validator(
    value: Any,
    context: ParameterValidationContext,
) -> bytes:
    match value:
        case bytes() as data:
            return data

        case str() as string:
            return b64decode(string)

        case _:
            raise TypeError(f"Expected 'str | bytes', received '{type(value).__name__}'")


def _b64_converter(
    value: bytes,
    /,
) -> str:
    return b64encode(value).decode("utf-8")



from asyncio import (
    CancelledError,
    Task,
)
from collections.abc import AsyncIterator
from typing import Protocol, Self, final, overload, runtime_checkable

from haiway import AsyncQueue, ctx

from draive.agents.idle import IdleMonitor
from draive.agents.node import Agent, AgentError, AgentMessage, AgentNode

__all__ = (
    "AgentRunner",
    "AgentRunnerOutput",
)


@runtime_checkable
class AgentRunnerOutput(Protocol):
    @overload
    def __call__(
        self,
        error: AgentError,
        /,
    ) -> None: ...

    @overload
    def __call__(
        self,
        messages: AgentMessage,
        /,
        *_messages: AgentMessage,
    ) -> None: ...

    def __call__(
        self,
        messages: AgentMessage | AgentError,
        /,
        *_messages: AgentMessage,
    ) -> None: ...


@final
class AgentRunner:
    @classmethod
    def run(
        cls,
        agent: AgentNode,
        /,
        output: AgentRunnerOutput,
        status: IdleMonitor,
    ) -> Self:
        queue: AsyncQueue[AgentMessage] = AsyncQueue()
        return cls(
            queue=queue,
            task=ctx.spawn(
                cls._draive,
                agent,
                queue,
                output,
                status,
            ),
            status=status,
        )

    def __init__(
        self,
        queue: AsyncQueue[AgentMessage],
        task: Task[None],
        status: IdleMonitor,
    ) -> None:
        self._queue: AsyncQueue[AgentMessage] = queue
        self._task: Task[None] = task
        self._status: IdleMonitor = status

    def send(
        self,
        message: AgentMessage,
        /,
    ) -> None:
        # not idle when received a message
        self._status.enter_task()
        self._queue.enqueue(message)

    @property
    def idle(self) -> bool:
        return self._status.idle

    async def wait_idle(self) -> None:
        await self._status.wait_idle()

    def cancel(self) -> None:
        self._task.cancel()

    def finish(self) -> None:
        self._queue.cancel()

    async def wait(self) -> None:
        await self._task

    async def finalize(self) -> None:
        self._queue.finish()
        await self._task

    @staticmethod
    async def _draive(
        agent: AgentNode,
        /,
        input: AsyncIterator[AgentMessage],  # noqa: A002
        output: AgentRunnerOutput,
        status: IdleMonitor,
    ) -> None:
        async with ctx.scope(agent.__str__()):  # pyright: ignore[reportDeprecated]
            agent_instance: Agent = agent.initialize()

            try:
                if agent.concurrent:
                    async for message in input:
                        task: Task[None] = ctx.spawn(
                            AgentRunner._handle,
                            message,
                            node=agent,
                            agent=agent_instance,
                            output=output,
                        )
                        task.add_done_callback(status.exit_task)

                else:
                    async for message in input:
                        try:
                            await AgentRunner._handle(
                                message,
                                node=agent,
                                agent=agent_instance,
                                output=output,
                            )

                        finally:
                            status.exit_task()

            except CancelledError:
                pass  # just end on cancel

            finally:  # cleanup status when exiting
                status.exit_all()

    @staticmethod
    async def _handle(
        message: AgentMessage,
        /,
        node: AgentNode,
        agent: Agent,
        output: AgentRunnerOutput,
    ) -> None:
        try:
            match await agent(message=message):
                case None:
                    pass  # nothing to do

                case [*results]:
                    output(*[result._with_sender(node) for result in results])  # pyright: ignore[reportPrivateUsage]

                case result:
                    output(result._with_sender(node))  # pyright: ignore[reportPrivateUsage]

        except CancelledError:
            pass  # ignore when Cancelled

        except Exception as exc:
            output(
                AgentError(
                    "Agent failed while processing the message",
                    agent=node,
                    message=message,
                    cause=exc,
                )
            )



from draive.guardrails.moderation import (
    GuardrailsInputModerationException,
    GuardrailsModeration,
    GuardrailsModerationChecking,
    GuardrailsModerationException,
    GuardrailsOutputModerationException,
)
from draive.guardrails.privacy import (
    GuardrailsAnonymization,
    GuardrailsAnonymizedContent,
    GuardrailsContentAnonymizing,
    GuardrailsContentDeanonymizing,
)
from draive.guardrails.quality import (
    GuardrailsQualityException,
    GuardrailsQualityVerification,
    GuardrailsQualityVerifying,
)

__all__ = (
    "GuardrailsAnonymization",
    "GuardrailsAnonymizedContent",
    "GuardrailsContentAnonymizing",
    "GuardrailsContentDeanonymizing",
    "GuardrailsInputModerationException",
    "GuardrailsModeration",
    "GuardrailsModerationChecking",
    "GuardrailsModerationException",
    "GuardrailsOutputModerationException",
    "GuardrailsQualityException",
    "GuardrailsQualityVerification",
    "GuardrailsQualityVerifying",
)



from draive.utils.memory import Memory
from draive.utils.processing import (
    Processing,
    ProcessingEvent,
    ProcessingState,
)
from draive.utils.rate_limit import RateLimitError
from draive.utils.splitting import split_sequence
from draive.utils.streams import ConstantStream, FixedStream
from draive.utils.vector_index import VectorIndex

__all__ = (
    "ConstantStream",
    "FixedStream",
    "Memory",
    "Processing",
    "ProcessingEvent",
    "ProcessingState",
    "RateLimitError",
    "VectorIndex",
    "split_sequence",
)



from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, Literal, Self, final, overload

from haiway import State, ctx

from draive.commons import META_EMPTY, Meta
from draive.instructions.file import InstructionsFileStorage
from draive.instructions.types import (
    Instruction,
    InstructionDeclaration,
    InstructionFetching,
    InstructionListFetching,
    InstructionMissing,
)
from draive.instructions.volatile import InstructionsVolatileStorage

__all__ = ("Instructions",)


async def _empty(
    name: str,
    **extra: Any,
) -> Sequence[InstructionDeclaration]:
    return ()


async def _none(
    name: str,
    /,
    *,
    arguments: Mapping[str, str | float | int] | None = None,
    **extra: Any,
) -> Instruction | None:
    return None


@final
class Instructions(State):
    @classmethod
    def of(
        cls,
        instruction: Instruction,
        *instructions: Instruction,
    ) -> Self:
        storage = InstructionsVolatileStorage((instruction, *instructions))

        return cls(
            list_fetching=storage.listing,
            fetching=storage.instruction,
        )

    @classmethod
    def file(
        cls,
        path: Path | str,
    ) -> Self:
        storage = InstructionsFileStorage(path=path)

        return cls(
            list_fetching=storage.listing,
            fetching=storage.instruction,
        )

    @classmethod
    async def fetch_list(
        cls,
        **extra: Any,
    ) -> Sequence[InstructionDeclaration]:
        return await ctx.state(cls).list_fetching(**extra)

    @overload
    @classmethod
    async def fetch(
        cls,
        instruction: InstructionDeclaration | str,
        /,
        *,
        arguments: Mapping[str, str | float | int] | None = None,
        **extra: Any,
    ) -> Instruction | None: ...

    @overload
    @classmethod
    async def fetch(
        cls,
        instruction: InstructionDeclaration | str,
        /,
        *,
        arguments: Mapping[str, str | float | int] | None = None,
        default: Instruction | str,
        **extra: Any,
    ) -> Instruction: ...

    @overload
    @classmethod
    async def fetch(
        cls,
        instruction: InstructionDeclaration | str,
        /,
        *,
        arguments: Mapping[str, str | float | int] | None = None,
        required: Literal[True],
        **extra: Any,
    ) -> Instruction: ...

    @classmethod
    async def fetch(
        cls,
        instruction: InstructionDeclaration | str,
        /,
        *,
        arguments: Mapping[str, str | float | int] | None = None,
        default: Instruction | str | None = None,
        required: bool = False,
        **extra: Any,
    ) -> Instruction | None:
        name: str = instruction if isinstance(instruction, str) else instruction.name

        if fetched := await ctx.state(Instructions).fetching(
            name,
            arguments=arguments,
            **extra,
        ):
            return fetched

        elif required and default is None:
            raise InstructionMissing(f"Missing instruction: '{name}'")

        elif isinstance(default, str):
            return Instruction.of(
                default,
                name=name,
                description=instruction.description
                if isinstance(instruction, InstructionDeclaration)
                else None,
                meta=None,
                arguments=arguments,
            )

        elif default is not None:
            return Instruction.of(
                default,
                arguments=arguments,
            )

        else:
            return None

    list_fetching: InstructionListFetching = _empty
    fetching: InstructionFetching = _none
    meta: Meta = META_EMPTY



from draive.generation.image import ImageGenerating, ImageGeneration
from draive.generation.model import (
    ModelGenerating,
    ModelGeneration,
    ModelGeneratorDecoder,
)
from draive.generation.text import TextGenerating, TextGeneration

__all__ = (
    "ImageGenerating",
    "ImageGeneration",
    "ModelGenerating",
    "ModelGeneration",
    "ModelGeneratorDecoder",
    "TextGenerating",
    "TextGeneration",
)



from draive.instructions.state import Instructions
from draive.instructions.template import InstructionTemplate, instruction
from draive.instructions.types import (
    Instruction,
    InstructionDeclaration,
    InstructionException,
    InstructionFetching,
    InstructionListFetching,
    InstructionMissing,
)

__all__ = (
    "Instruction",
    "InstructionDeclaration",
    "InstructionException",
    "InstructionFetching",
    "InstructionListFetching",
    "InstructionMissing",
    "InstructionTemplate",
    "Instructions",
    "instruction",
)



from typing import cast

from draive.evaluation import EvaluationScore, EvaluationScoreValue, evaluator
from draive.multimodal import Multimodal, MultimodalContent, MultimodalTagElement
from draive.stages import Stage

__all__ = ("conciseness_evaluator",)


INSTRUCTION: str = """\
You are evaluating the provided content according to the defined criteria.

<INSTRUCTION>
Compare the REFERENCE and the EVALUATED content by carefully examining them, then rate\
 the EVALUATED content using solely a conciseness metric according to the EVALUATION_CRITERIA.
Think step by step and provide explanation of the score before the final score.
Use the explained RATING scale and the requested FORMAT to provide the result.
</INSTRUCTION>

<EVALUATION_CRITERIA>
Evaluated metric is conciseness - a extent to which the EVALUATED content is brief and to the\
 point while still covering all key information.
A concise content avoids unnecessary details and repetition, also avoiding being overly verbose\
 or include irrelevant information.
</EVALUATION_CRITERIA>
{guidelines}
<RATING>
Assign a conciseness score using exact name of one of the following values:
- "poor" is very low conciseness, the content is excessively verbose with much irrelevant\
 information.
- "fair" is low conciseness, the content contains unnecessary details and some irrelevant\
 information.
- "good" is moderate conciseness, the content is somewhat concise but could be more focused.
- "excellent" is high conciseness, the content is mostly concise with minimal unnecessary\
 information.
- "perfect" is very high conciseness, the content is highly concise, containing only essential\
 information.
Use the "none" value for content that cannot be rated at all.
</RATING>

<FORMAT>
The final result containing only the numerical score value HAVE to be put inside a `RESULT`\
 xml tag within the result i.e. `<RESULT>good</RESULT>`.
</FORMAT>
"""


@evaluator(name="conciseness")
async def conciseness_evaluator(
    evaluated: Multimodal,
    /,
    *,
    reference: Multimodal,
    guidelines: str | None = None,
) -> EvaluationScore:
    if not evaluated:
        return EvaluationScore(
            value=0,
            comment="Input was empty!",
        )

    if not reference:
        return EvaluationScore(
            value=0,
            comment="Reference was empty!",
        )

    completion: MultimodalContent = await Stage.completion(
        MultimodalContent.of(
            "<REFERENCE>",
            reference,
            "</REFERENCE>\n<EVALUATED>",
            evaluated,
            "</EVALUATED>",
        ),
        instruction=INSTRUCTION.format(
            guidelines=f"\n<GUIDELINES>\n{guidelines}\n</GUIDELINES>\n"
            if guidelines is not None
            else ""
        ),
    ).execute()

    if result := MultimodalTagElement.parse_first(
        "RESULT",
        content=completion,
    ):
        return EvaluationScore.of(
            cast(EvaluationScoreValue, result.content.to_str()),
            comment=completion.to_str(),
        )

    else:
        raise ValueError("Invalid evaluator result:\n%s", completion)



from typing import Any

from google.genai import Client
from google.genai.client import HttpOptionsDict  # pyright: ignore[reportPrivateImportUsage]
from haiway import getenv_str

__all__ = ("GeminiAPI",)


class GeminiAPI:
    __slots__ = ("_client",)

    def __init__(
        self,
        api_key: str | None,
        http_options: HttpOptionsDict | None,
        **extra: Any,
    ) -> None:
        self._client = Client(
            api_key=api_key or getenv_str("GEMINI_API_KEY"),
            http_options=http_options,
            **extra,
        )



import json
from collections.abc import Callable, Iterable
from typing import Any, Literal, cast

from haiway import Missing, not_missing
from openai import NOT_GIVEN, NotGiven
from openai.types.chat import (
    ChatCompletionContentPartParam,
    ChatCompletionMessageParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolParam,
)
from openai.types.chat.completion_create_params import ResponseFormat
from openai.types.shared_params.response_format_json_schema import ResponseFormatJSONSchema

from draive.lmm import (
    LMMCompletion,
    LMMContextElement,
    LMMInput,
    LMMOutputSelection,
    LMMToolRequests,
    LMMToolResponses,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.multimodal import (
    MediaData,
    MediaReference,
    MultimodalContent,
    MultimodalContentElement,
    TextContent,
)
from draive.parameters import DataModel

__all__ = (
    "content_element_as_content_part",
    "context_element_as_messages",
    "tools_as_tool_config",
)


def content_element_as_content_part(
    element: MultimodalContentElement,
    vision_details: Literal["auto", "low", "high"] | Missing,
) -> ChatCompletionContentPartParam:
    match element:
        case TextContent() as text:
            return {
                "type": "text",
                "text": text.text,
            }

        case MediaData() as media_data:
            if media_data.kind != "image":
                raise ValueError("Unsupported message content", media_data)

            return {
                "type": "image_url",
                "image_url": {
                    "url": media_data.to_data_uri(safe_encoding=False),
                    "detail": cast(Literal["auto", "low", "high"], vision_details)
                    if not_missing(vision_details)
                    else "auto",
                },
            }

        case MediaReference() as media_reference:
            if media_reference.kind != "image":
                raise ValueError("Unsupported message content", media_reference)

            return {
                "type": "image_url",
                "image_url": {
                    "url": media_reference.uri,
                    "detail": cast(Literal["auto", "low", "high"], vision_details)
                    if not_missing(vision_details)
                    else "auto",
                },
            }

        case DataModel() as data:
            return {
                "type": "text",
                "text": data.to_json(),
            }


def context_element_as_messages(
    element: LMMContextElement,
    /,
    vision_details: Literal["auto", "low", "high"] | Missing,
) -> Iterable[ChatCompletionMessageParam]:
    match element:
        case LMMInput() as input:
            return (
                {
                    "role": "user",
                    "content": [
                        content_element_as_content_part(
                            element=element,
                            vision_details=vision_details,
                        )
                        for element in input.content.parts
                    ],
                },
            )

        case LMMCompletion() as completion:
            return (
                {
                    "role": "assistant",
                    # TODO: OpenAI models generating media?
                    "content": completion.content.to_str(),
                },
            )

        case LMMToolRequests() as tool_requests:
            return (
                {
                    "role": "assistant",
                    "tool_calls": [
                        {
                            "id": request.identifier,
                            "type": "function",
                            "function": {
                                "name": request.tool,
                                "arguments": json.dumps(dict(request.arguments)),
                            },
                        }
                        for request in tool_requests.requests
                    ],
                },
            )

        case LMMToolResponses() as tool_responses:
            return (
                {
                    "role": "tool",
                    "tool_call_id": response.identifier,
                    "content": response.content.to_str(),
                }
                for response in tool_responses.responses
            )


def output_as_response_declaration(
    output: LMMOutputSelection,
) -> tuple[
    ResponseFormat | ResponseFormatJSONSchema | NotGiven,
    Callable[[MultimodalContent], MultimodalContent],
]:
    match output:
        case "auto":
            return (
                NOT_GIVEN,
                _auto_output_conversion,
            )

        case ["text"] | "text":
            return (
                {"type": "text"},
                _auto_output_conversion,
            )

        case "json":
            return (
                {"type": "json_object"},
                _json_output_conversion,
            )

        case "image":
            raise NotImplementedError("image output is not supported yet")

        case "audio":
            raise NotImplementedError("audio output is not supported yet")

        case "video":
            raise NotImplementedError("video output is not supported yet")

        case [*_]:
            raise NotImplementedError("multimodal output is not supported yet")

        case model:
            return (
                {
                    "type": "json_schema",
                    "json_schema": {
                        "name": model.__name__,
                        "schema": cast(
                            dict[str, Any],
                            model.__PARAMETERS_SPECIFICATION__,
                        ),
                        "strict": False,
                    },
                },
                _prepare_model_output_conversion(model),
            )


def _auto_output_conversion(
    output: MultimodalContent,
    /,
) -> MultimodalContent:
    return output


def _text_output_conversion(
    output: MultimodalContent,
    /,
) -> MultimodalContent:
    return MultimodalContent.of(output.to_str())


def _audio_output_conversion(
    output: MultimodalContent,
    /,
) -> MultimodalContent:
    return MultimodalContent.of(*output.media("audio"))


def _json_output_conversion(
    output: MultimodalContent,
    /,
) -> MultimodalContent:
    return MultimodalContent.of(DataModel.from_json(output.to_str()))


def _prepare_model_output_conversion(
    model: type[DataModel],
    /,
) -> Callable[[MultimodalContent], MultimodalContent]:
    def _model_output_conversion(
        output: MultimodalContent,
        /,
    ) -> MultimodalContent:
        return MultimodalContent.of(model.from_json(output.to_str()))

    return _model_output_conversion


def tool_specification_as_tool(
    tool: LMMToolSpecification,
    /,
) -> ChatCompletionToolParam:
    return {
        "type": "function",
        "function": {
            "name": tool["name"],
            "description": tool["description"] or "",
            "parameters": cast(dict[str, Any], tool["parameters"])
            or {"type": "object", "properties": {}},
        },
    }


def tools_as_tool_config(
    tools: Iterable[LMMToolSpecification] | None,
    /,
    *,
    tool_selection: LMMToolSelection,
) -> tuple[
    ChatCompletionToolChoiceOptionParam | NotGiven, list[ChatCompletionToolParam] | NotGiven
]:
    tools_list: list[ChatCompletionToolParam] = [
        tool_specification_as_tool(tool) for tool in (tools or [])
    ]
    if not tools_list:
        return (NOT_GIVEN, NOT_GIVEN)

    match tool_selection:
        case "auto":
            return ("auto", tools_list)

        case "none":
            return (NOT_GIVEN, NOT_GIVEN)

        case "required":
            return ("required", tools_list)

        case tool:
            assert tool in (tools or []), "Can't suggest a tool without using it"  # nosec: B101

            return (
                {
                    "type": "function",
                    "function": {"name": tool["name"]},
                },
                tools_list,
            )



from collections.abc import Callable, Sequence
from typing import Any, cast, overload

from haiway import AttributePath, State, ctx

from draive.embedding.types import Embedded, ValueEmbedding
from draive.parameters import DataModel

__all__ = (
    "ImageEmbedding",
    "TextEmbedding",
)


class TextEmbedding(State):
    @overload
    @classmethod
    async def embed(
        cls,
        content: str,
        /,
        **extra: Any,
    ) -> Embedded[str]: ...

    @overload
    @classmethod
    async def embed(
        cls,
        content: Sequence[str],
        /,
        **extra: Any,
    ) -> Sequence[Embedded[str]]: ...

    @overload
    @classmethod
    async def embed[Value: DataModel | State](
        cls,
        content: Value,
        /,
        attribute: Callable[[Value], str] | AttributePath[Value, str] | str,
        **extra: Any,
    ) -> Embedded[Value]: ...

    @overload
    @classmethod
    async def embed[Value: DataModel | State](
        cls,
        content: Sequence[Value],
        /,
        attribute: Callable[[Value], str] | AttributePath[Value, str] | str,
        **extra: Any,
    ) -> Sequence[Embedded[Value]]: ...

    @classmethod
    async def embed[Value: DataModel | State | str](
        cls,
        content: Sequence[Value] | Sequence[str] | Value | str,
        /,
        attribute: Callable[[Value], str] | AttributePath[Value, str] | str | None = None,
        **extra: Any,
    ) -> Sequence[Embedded[Value]] | Sequence[Embedded[str]] | Embedded[Value] | Embedded[str]:
        str_selector: Callable[[Value], str] | None
        match attribute:
            case None:
                str_selector = None

            case Callable() as function:  # pyright: ignore[reportUnknownVariableType]
                str_selector = function

            case path:
                assert isinstance(  # nosec: B101
                    path, AttributePath
                ), "Prepare parameter path by using Self._.path.to.property"
                str_selector = cast(AttributePath[Value, str], path).__call__

        match content:
            case [*elements]:
                return await ctx.state(cls).embedding(
                    elements,
                    attribute=str_selector,
                    **extra,
                )

            case element:
                return (
                    await ctx.state(cls).embedding(
                        [element],
                        attribute=str_selector,
                        **extra,
                    )
                )[0]

    embedding: ValueEmbedding[Any, str]


class ImageEmbedding(State):
    @overload
    @classmethod
    async def embed(
        cls,
        content: bytes,
        /,
        **extra: Any,
    ) -> Embedded[bytes]: ...

    @overload
    @classmethod
    async def embed(
        cls,
        content: Sequence[bytes],
        /,
        **extra: Any,
    ) -> Sequence[Embedded[bytes]]: ...

    @overload
    @classmethod
    async def embed[Value: DataModel | State](
        cls,
        content: Value,
        /,
        attribute: Callable[[Value], bytes] | AttributePath[Value, bytes] | bytes,
        **extra: Any,
    ) -> Embedded[Value]: ...

    @overload
    @classmethod
    async def embed[Value: DataModel | State](
        cls,
        content: Sequence[Value],
        /,
        attribute: Callable[[Value], bytes] | AttributePath[Value, bytes] | bytes,
        **extra: Any,
    ) -> Sequence[Embedded[Value]]: ...

    @classmethod
    async def embed[Value: DataModel | State | bytes](
        cls,
        content: Sequence[Value] | Sequence[bytes] | Value | bytes,
        /,
        attribute: Callable[[Value], bytes] | AttributePath[Value, bytes] | bytes | None = None,
        **extra: Any,
    ) -> Sequence[Embedded[Value]] | Sequence[Embedded[bytes]] | Embedded[Value] | Embedded[bytes]:
        bytes_selector: Callable[[Value], bytes] | None
        match attribute:
            case None:
                bytes_selector = None

            case Callable() as function:  # pyright: ignore[reportUnknownVariableType]
                bytes_selector = function

            case path:
                assert isinstance(  # nosec: B101
                    path, AttributePath
                ), "Prepare parameter path by using Self._.path.to.property"
                bytes_selector = cast(AttributePath[Value, bytes], path).__call__

        match content:
            case [*elements]:
                return await ctx.state(cls).embedding(
                    elements,
                    attribute=bytes_selector,
                    **extra,
                )

            case element:
                return (
                    await ctx.state(cls).embedding(
                        [element],
                        attribute=bytes_selector,
                        **extra,
                    )
                )[0]

    embedding: ValueEmbedding[Any, bytes]
